from ..base_executor import BaseExecutor
from ...insight import InsightState, StrategyDependantConfirmation
from ....broker.interfaces import IOrderSide

class PercentageDCALevels(BaseExecutor):
    """
    ### Executor for Adding DCA Levels
    This executor is used to add DCA levels to a new insight that already had a desired quantity.

    It will append the DCA levels as childens to the insight based on the DCA percentage and levels set in the alpha.
    The child insights will have the same TP as the parent insight bur will alter the SL below / above  the last DCA level.
    The children insight strategyType will be set to the alpha name with a suffix of "_CHILD".

    :param strategy (BaseStrategy): The strategy instance
    :param includes (list[str]): The alphas to include
    :param dcaPercentage (float): The percentage to space the DCA levels
    :param dcalevels (int): The number of DCA levels to add
    Author:
        @isaac-diaby
    """
    includes: list[str]
    """alpha to include"""
    dcaPercentage: float
    dcalevels: int


    def __init__(self, strategy, includes: list[str], dcaPercentage: float = 0.01, dcalevels: int = 5):
        super().__init__(strategy, InsightState.NEW, "1.0")
        assert dcaPercentage > 0, "DCA percentage must be greater than 0."
        assert dcalevels > 0, "DCA levels must be greater than 0."
        self.includes = includes

        self.dcaPercentage = dcaPercentage
        self.dcalevels = dcalevels

    def run(self, insight):
        # check if insight strategyType alpha is part if the includes list
        if self.includes and insight.strategyType not in self.includes:
            return self.returnResults(True, True, "Alpha is not part of the includes list.")
        # Check if the insight has a desired quantity
        if insight.quantity is None:
            return self.returnResults(True, True, "Insight does not have a desired quantity.")
        # Check if the insight is a parent insight
        if insight.PARANT:
            return self.returnResults(True, True, "Insight is not a parent insight.")
        try:
            latestBar = self.get_latest_bar(insight.symbol)
            desiredQuantity = insight.quantity
            entry = insight.limit_price if insight.limit_price else latestBar['close']
            SL = insight.SL
            TP = insight.TP

            price_levels = [] # List to store the price levels for DCA

            price_levels.append(entry) # Append the entry price

            for i in range(1, self.dcalevels+1):
                # Every X% above / below the entry price
                if insight.side == IOrderSide.BUY:
                    dca_level = self.STRATEGY.tools.dynamic_round(entry * (1 - (self.dcaPercentage*i)), insight.symbol)
                else: # shorting
                    dca_level = self.STRATEGY.tools.dynamic_round(entry * (1 + (self.dcaPercentage*i)), insight.symbol)
                price_levels.append(dca_level)

            quantities = self.calculate_dca_quantities(desiredQuantity, price_levels)
            # the first quantity is the parent insight quantity
            insight.update_quantity(quantities[0])
            dca_SL = price_levels[-1] - ((entry - SL)/2)
            # Update the SL of the parent insight
            insight.SL = dca_SL


            # Add the DCA levels as children insights
            for i in range(1, len(quantities)-1):
                childInsight = insight.addChildInsight(
                    side=insight.side,
                    quantity=quantities[i] if insight.side == IOrderSide.BUY else quantities[len(quantities)-i],
                    limit_price=price_levels[i],
                    SL=dca_SL,
                    TP=TP,
                    executionDepends=[
                        StrategyDependantConfirmation.UPSTATE if insight.side == IOrderSide.BUY else StrategyDependantConfirmation.DOWNSTATE],
                )
            
            drawdowns, total_drawdowns = self.calculate_drawdowns(price_levels, quantities, dca_SL)
            print("Drawdowns: ", drawdowns)
            print("Total Drawdowns: ", total_drawdowns)
            return self.returnResults(True, True, "Added DCA levels to insight.")
        except Exception as e:
            return self.returnResults(False, False, f"Error Adding DCA to insight: {e}")
        
    def calculate_dca_quantities(self, total_quantity, price_levels):
        """
        Calculate the quantity to buy at each DCA level using the Inverse Price Proportionality method.

        :param total_quantity: Total desired quantity to accumulate.
        :param price_levels: List of prices at each DCA level.
        :return: List of quantities to buy at each level.
        """
        # Calculate the denominator (sum of inverses of prices)
        denominator = sum(1 / price for price in price_levels)
        
        # Calculate the quantity to buy at each level
        quantities = [round((total_quantity / denominator) * (1 / price), 4) for price in price_levels]
        
        return quantities
    def calculate_drawdowns(self, entry_prices, quantities, stop_loss_price):
        """
        Calculate the expected drawdown at each DCA level.
        
        :param entry_prices: List of entry prices at each DCA level.
        :param quantities: List of quantities bought at each DCA level.
        :param stop_loss_price: The stop loss price.
        :return: List of drawdowns for each level and total drawdown up to each level.
        """
        drawdowns = []
        total_drawdowns = []
        total_quantity = 0
        weighted_average_price = 0

        for i in range(len(entry_prices)):
            price = entry_prices[i]
            quantity = quantities[i]
            
            # Calculate drawdown for this level
            drawdown = (price - stop_loss_price) * quantity
            drawdowns.append(drawdown)
            
            # Update total quantity and weighted average price
            total_quantity += quantity
            weighted_average_price = ((weighted_average_price * (total_quantity - quantity)) + (price * quantity)) / total_quantity
            
            # Calculate total drawdown up to this level
            total_drawdown = (weighted_average_price - stop_loss_price) * total_quantity
            total_drawdowns.append(total_drawdown)

        return drawdowns, total_drawdowns


from ..base_executor import BaseExecutor
from ...insight import InsightState
from ....broker.interfaces import IOrderType


class MarketOrderEntryPriceExecutor(BaseExecutor):
    """
    ### Executor for Market Order Entry Price
    This executor is used to set the entry price for market orders. The entry price is set to the current close price of the asset.
    Get price from latest bar if limit price is not set.

    :param strategy (BaseStrategy): The strategy instance

    Note: This is only applicable for insights that are set to market orders and have no limit price set. 
    If a limit is set, executor passes the insight to the next executor.

    Author:
        @isaac-diaby
    """

    def __init__(self, strategy):
        super().__init__(strategy, InsightState.NEW, "1.0")

    def run(self, insight):
        if insight.type != IOrderType.MARKET:
            return self.returnResults(True, True, "Insight already has a limit price set. Passing to next executor.")
        try:
            # Set the limit price to the current close price
            latestBar = self.get_latest_bar(insight.symbol)
            self.STRATEGY.insights[insight.INSIGHT_ID].update_limit_price(
                latestBar.close)
            return self.returnResults(True, True, f"limit price set to current close price: {latestBar.close}")
        except Exception as e:
            return self.returnResults(False, False, f"Error setting limit price: {e}")



from dash import html


def PlaceholderChart(title: str, children=None):
    return html.Div(
        className='bg-primary-foreground p-6 rounded-lg shadow-lg border border-accent',
        children=[
            html.H3(
                className='text-xl font-semibold text-accent mb-4',
                children=title
            ),
            (html.Div(
                className='reletive min-h-64 bg-primary-light rounded flex items-center justify-center',
                children=[
                    html.P(
                        className='text-accent',
                        children='Chart Placeholder'
                    )
                ]
            ) if children is None else html.Div(
                className='reletive overflow-hidden  min-h-64 rounded flex items-center justify-center',
                children=children
            ))
        ]
    )




from dash import html, dcc, callback, register_page, no_update
from dash.dependencies import Input, Output, State

from OlympusTrader.ui.components.dashboardNav import subnav
from OlympusTrader.ui.interfaces.store import STRATEGY_SYNC_MAPPING

register_page(
    __name__, title="OlympusTrader - Settings", path='/settings')

layout = html.Div([
    subnav(),
    html.Div([

        html.H3("Settings 🚧", className="text-xl font-semibold text-accent mb-2"),
        html.P('This page is under construction. Please check back later.', className= "mb-4"),

        html.Div([
            # Strategy Sync Settings
            html.Div([
                html.Div([
                    html.H2('Strategy Sync Settings',
                            className="text-2xl font-semibold text-accent mb-4"),
                    html.Div([
                        html.Div([
                            html.Label(
                                'Sync Rate (seconds)', className="block text-sm font-medium text-accent mb-2"),
                            html.Div([
                                dcc.Input(
                                    id='sync-rate-input',
                                    type='number',
                                    min=1,
                                    className="bg-primary-light border border-accent text-white rounded-md px-4 py-2 flex-grow focus:outline-none focus:ring-2 focus:ring-accent"
                                ),
                                html.Button(
                                    'Apply',
                                    id='sync-rate-submit',
                                    className="px-4 py-2 bg-accent text-black rounded hover:bg-[#a88c5b] transition-colors flex items-center"
                                )
                            ], className="flex space-x-3"),
                            html.P('Current sync rate: ', id='current-sync-rate',
                                   className="text-sm text-gray-400 mt-2")
                        ])
                    ], className="space-y-4")
                ], className="bg-primary-foreground rounded-lg shadow-lg border border-accent p-6 mb-6")
            ], className="lg:col-span-2")
        ], className="grid grid-cols-1 lg:grid-cols-3 gap-8")
    ], className="flex-grow container mx-auto py-8 text-white")
])


@callback(
    [
        Output(STRATEGY_SYNC_MAPPING.id, 'n_intervals'),
        Output('current-sync-rate', 'children'),
    ],
    Input('sync-rate-submit', 'n_clicks'),
    [
        State('sync-rate-input', 'value'),
        State(STRATEGY_SYNC_MAPPING.id, 'n_intervals')
    ],
    prevent_initial_call=True
)
def update_sync_rate(n_clicks, value, current_n_intervals):
    if value is None:
        return no_update, f'Current sync rate: {(current_n_intervals/1000)} seconds'
    if value < 1:
        return no_update
    
    print(f"Updating sync rate from {current_n_intervals} to {value} seconds")

    interval = value * 1000  # Convert to milliseconds
    return interval, f'Current sync rate: {value} seconds'



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from types import NoneType
from typing import List, Literal, TYPE_CHECKING, Optional, Self, Type, Union
from uuid import uuid4, UUID
from numpy import isnan

from ..utils.timeframe import ITimeFrame
from ..broker.interfaces import IAsset, IOrderSide, IOrderType, IOrderClass, IOrderLegs, IOrderLeg
from ..strategy.interfaces import IStrategyMode


if TYPE_CHECKING:
    from ..broker.base_broker import BaseBroker


def get_BaseBroker():
    from ..broker.base_broker import BaseBroker
    return BaseBroker


class StrategyTypes(Enum):
    MANUAL = 'MANUAL'
    TEST = 'TESTING'


class StrategyDependantConfirmation(Enum):
    NONE = 'NONE'
    HRVCM = 'High Relative Volume Confirmation Model'
    LRVCM = 'Low Relative Volume Confirmation Model'
    HTFCM = 'High Time Frame Confirmation Model'
    LTFCM = 'Low Time Frame Confirmation Model'
    HCCM = 'High Confidence Confirmation Model'
    LCCM = 'Low Confidence Confirmation Model'
    UPSTATE = 'Up State Confirmation Model'
    DOWNSTATE = 'Down State Confirmation Model'
    FLATSTATE = 'Flat State Confirmation Model'


class InsightState(Enum):
    NEW = 'NEW'
    EXECUTED = 'EXECUTED'
    FILLED = 'FILLED'
    CLOSED = 'CLOSED'
    CANCELED = 'CANCELED'
    REJECTED = 'REJECTED'  # EXPIRED = 'EXPIRED'


class PartialCloseResult:
    order_id: str
    side: IOrderSide
    quantity: float
    entry_price: float
    filled_price: float

    def __init__(self, order_id: str, side: IOrderSide, quantity: float, price: float):
        self.order_id = order_id
        self.side = side
        self.quantity = quantity
        self.entry_price = price

    def __str__(self):
        return f"Partial Close: {self.order_id} - {self.quantity} @ {self.entry_price} - Filled At: {self.filled_price if self.filled_price else 'N/A'}"

    def set_filled_price(self, filled_price: float):
        self.filled_price = filled_price

    def getPL(self):
        assert self.filled_price != None, 'Filled price is not set'
        if self.side == IOrderSide.BUY:
            return round((self.filled_price - self.entry_price) * self.quantity, 2)
        else:
            return round((self.entry_price - self.filled_price) * self.quantity, 2)


class Insight:
    INSIGHT_ID: UUID
    """Unique Insight ID"""
    PARANT: Optional[UUID] = None
    """Parent Insight ID"""
    CHILDREN: dict[UUID, Self]

    order_id: str = None
    """Main Order ID for the Insight that entered the position"""
    side: IOrderSide = None  # buy or sell
    opposite_side: IOrderSide = None
    symbol: str = None  # symbol to trade
    quantity: Optional[float] = None  # quantity to trade
    contracts: Optional[float] = None  # quantity to trade in contracts

    type: IOrderType = None  # market, limit, stop, stop_limit, trailing_stop
    classType: IOrderClass = None  # simple, bracket, oco, oto
    limit_price: Optional[float] = None
    """Entry Price"""
    TP: Optional[List[float]] = None
    """Take profit levels - The last index is the final take profit level"""
    SL: Optional[float] = None
    """Stop loss level"""
    strategyType: Optional[Union[StrategyTypes, str]] = None  # strategy type
    confidence: float = None  # confidence in insight
    tf: ITimeFrame = None  # timeframe
    periodUnfilled: Optional[int] = None  # time to live when unfilled
    # predicted time to live when opened to reach take profit
    periodTillTp: Optional[int] = None
    executionDepends: List[StrategyDependantConfirmation] = [
        StrategyDependantConfirmation.NONE]  # execution depends on
    state: InsightState = InsightState.NEW
    """State of the Insight"""
    createAt: datetime = datetime.now()
    updatedAt: datetime = datetime.now()
    filledAt: Optional[datetime] = None
    closedAt: Optional[datetime] = None
    close_order_id = None
    """Close Order ID for the Insight that closed the position"""
    partial_closes: list[PartialCloseResult] = []
    legs: IOrderLegs = IOrderLegs(
        take_profit=None, stop_loss=None, trailing_stop=None)
    close_price: Optional[float] = None  # price to close at

    marketChanged: bool = False
    _cancelling: bool = False
    """Flag to check if the insight is being canceled"""
    _closing: bool = False
    """Flag to check if the insight is being closed"""
    _partial_filled_quantity: Optional[float] = None

    MODE: IStrategyMode = IStrategyMode.LIVE
    BROKER: get_BaseBroker = None
    ASSET: IAsset = None

    def __init__(self, side: IOrderSide, symbol: str,  strategyType: Union[StrategyTypes, str], tf: ITimeFrame, quantity: Optional[float] = 1, limit_price: Optional[float] = None, TP: Optional[List[float]] = None, SL: Optional[float] = None,  confidence: float = 0.1, executionDepends: List[StrategyDependantConfirmation] = [StrategyDependantConfirmation.NONE], periodUnfilled: int = 2, periodTillTp: int = 10,  parent: Optional[UUID] = None):
        assert side in IOrderSide, 'Invalid Order Side'
        self.INSIGHT_ID = uuid4()
        if parent:
            self.PARANT = parent  # Parent Insight ID
        self.CHILDREN = {}

        self.side = side  # buy or sell
        self.opposite_side = IOrderSide.BUY if (
            self.side == IOrderSide.SELL) else IOrderSide.SELL
        self.symbol = symbol  # symbol to trade
        self.quantity = quantity  # quantity to trade
        self.limit_price = limit_price  # price to enter at
        self.strategyType = strategyType  # strategy type
        # confidence in insight clamp between 0.01 and 1
        self.confidence = min(max(confidence, 0.01), 1)
        self.TP = TP  # take profit levels
        self.SL = SL  # stop loss
        self.tf = tf  # timeframe
        self.periodUnfilled = periodUnfilled  # time to live when unfilled
        # predicted time to live when opened to reach take profit
        self.periodTillTp = periodTillTp
        self.executionDepends = executionDepends  # execution depends on
        # self.state = InsightState.NEW
        self.createAt = datetime.now()
        self.updatedAt = datetime.now()

        if self.limit_price == None:
            self.type = IOrderType.MARKET
        else:
            self.type = IOrderType.LIMIT

        if self.TP and self.SL:
            self.classType = IOrderClass.BRACKET
        else:
            self.classType = IOrderClass.SIMPLE

    def __str__(self):
        if self.strategyType == StrategyTypes.MANUAL:
            return f"Insight - {self.state:<5} : {self.strategyType:^16} - {self.symbol:^8} :: {self.side:^5}: {str(self.quantity)} @ MARKET"
        return f"Insight - {self.state:<5} : {self.strategyType:^16} - {self.symbol:^8} :: {self.side:^5}: {str(self.quantity)} @ {str(self.limit_price if self.limit_price else "MARKET"):^5} - TP: {str(self.TP):^5} - SL: {self.SL:^5} - Ratio: {str(self.getPnLRatio()):^10} - TTLUF/TTL: {str(self.periodUnfilled):^5}/{str(self.periodTillTp):^5} - UDA: {self.updatedAt}"

    def submit(self, rejectInvalid: bool = True, partialCloseInsight=None, closeInsight=False):

        if self.state == InsightState.NEW or self.state == InsightState.FILLED or self.state == InsightState.EXECUTED:
            try:
                if partialCloseInsight != None:
                    # Submit a partial close order
                    order = self.BROKER.execute_insight_order(
                        partialCloseInsight, self.ASSET)
                elif closeInsight:
                    # Submit a close order for the insight

                    #  check if the insight has a take profit or stop loss order leg that need to be canceled before closing the position
                    if self.takeProfitOrderLeg:
                        self.cancelTakeProfitLeg()
                    if self.stopLossOrderLeg:
                        self.cancelStopLossLeg()
                    # close insight
                    closeInsight = Insight(
                        self.opposite_side, self.symbol, StrategyTypes.MANUAL, self.tf, self.quantity)
                    closeInsight.order_id = self.order_id
                    # Submit the insight to close the position
                    order = self.BROKER.execute_insight_order(
                        closeInsight, self.ASSET)

                else:
                    # Submit the insight to enter a position
                    order = self.BROKER.execute_insight_order(self, self.ASSET)

                if order:
                    if self.state == InsightState.NEW:
                        self.updateOrderID(
                            order['order_id'])
                        if order["legs"]:
                            self.updateLegs(order["legs"])
                        self.updateState(
                            InsightState.EXECUTED, f"Order ID: {order['order_id']}")

                    elif self.state == InsightState.FILLED:
                        # Add To Patial Close
                        if partialCloseInsight != None:
                            self.partial_closes.append(
                                PartialCloseResult(order['order_id'], self.side, partialCloseInsight.quantity, self.limit_price))
                            self.quantity -= partialCloseInsight.quantity
                        else:
                            # Close the position
                            self.updateCloseOrderID(order['order_id'])
                            # Strategy should handle the incoming closing of the position
                            # self.updateState(
                            #     InsightState.CLOSED, f"Close Order ID: {closeOrder['order_id']}")

                    return True
                else:
                    if self.state == InsightState.NEW:
                        self.updateState(
                            InsightState.REJECTED, 'Failed to place order')

            except BaseException as error:
                if self.state == InsightState.NEW:
                    self.updateState(InsightState.REJECTED, f"Error: {error}")
                else:
                    # Close function should handle the error if it was called by the close function
                    raise error

        else:
            print("Insight is not in a valid state to be submitted")
            if rejectInvalid:
                self.updateState(InsightState.REJECTED,
                                 'Invalid Entry Insight')

        return False

    def cancel_order_by_id(self, order_id: str):
        if order_id == None or order_id == "":
            print("Order ID is not set")
            return False
        try:
            order = self.BROKER.cancel_order(order_id)
            if order:
                return order
            else:
                return False
        except BaseException as e:
            raise e

    def cancel(self):
        if self.state == InsightState.EXECUTED:
            try:
                if self._cancelling:
                    if self.state == InsightState.FILLED:
                        return False
                    return True
                order = self.cancel_order_by_id(self.order_id)
                if order:
                    self._cancelling = True
                    return True

            except BaseException as e:
                if e.args[0]["code"] == "already_filled":
                    self._cancelling = False
                    # self.updateState(InsightState.FILLED,
                    #                  'Already filled Trade')
                    # return True
                    return False
        if self.state == InsightState.NEW:
            self.updateState(InsightState.REJECTED, 'Before Executed Canceled')
            return True
        if self._cancelling == False:
            print("Insight is not in a valid state to be canceled",self.state, self.INSIGHT_ID)
        return False

    def close(self, quantity: Optional[float] = None, retry: bool = True, bypassStateCheck: bool = False):
        """
        Used to close a Insight (position) or to partially close a position (if you have multiple take profit levels)
        """
        if (self.state == InsightState.FILLED or bypassStateCheck) and (self._closing == False):
            partialClose = False
            try:
                if quantity != None and self.quantity:
                    if quantity < self.quantity:
                        partialClose = True
                    else:
                        # Close 100% of the position if the quantity is greater than the insight's position size
                        quantity = self.quantity
                if partialClose == False:
                    # Close 100% of the position for this insight
                    closeOrder = self.submit(closeInsight=True)
                    if closeOrder:
                        self._closing = True
                        return True
                else:
                    partialCloseInsight = Insight(
                        self.opposite_side, self.symbol, StrategyTypes.MANUAL, self.tf, quantity)
                    # This is needed for MT5 as you need to link the position ID to the order. We would also use this on the broker end to determain id the order is being closed (force close!)
                    partialCloseInsight.order_id = self.order_id
                    closePartialOrder = self.submit(
                        partialCloseInsight=partialCloseInsight)
                    if closePartialOrder:
                        if self.TP:
                            self.TP.pop(0)  # Remove the first TP level
                        return True

            except BaseException as e:
                if e.args[0]["code"] == "insufficient_balance":
                    # '{"available":"0.119784","balance":"0.119784","code":40310000,"message":"insufficient balance for BTC (requested: 0.12, available: 0.119784)","symbol":"USD"}'
                    if e.args[0]["data"].get("balance") != None:
                        holding = float(e.args[0]["data"]["balance"])
                        if (holding > 0):
                            # Close 100% of the position
                            self.quantity = abs(holding)
                            # Retry closing the position once
                            if retry:
                                return self.close(retry=False)
                            else:
                                print(f"failed to close position: {
                                    holding} remaining!")
                        else:
                            # The position has already been closed as the balance is 0
                            self.updateState(
                                InsightState.CANCELED, f"No funds to close position")
        elif self._closing == False:
            print("Insight is not in a valid state to be closed", self.state)

        return False

    def addChildInsight(self, side: IOrderSide, quantity: Optional[float] = 1, limit_price: Optional[float] = None, TP: Optional[List[float]] = None, SL: Optional[float] = None, executionDepends: List[StrategyDependantConfirmation] = [StrategyDependantConfirmation.NONE], periodUnfilled: int = None, periodTillTp: int = None) -> Self:
        """Add a child insight to the parent insight. The child insight will be executed after the parent insight is filled."""
        childInsight = Insight(
            parent=self.INSIGHT_ID,
            symbol=self.symbol,
            strategyType=self.strategyType+'-CHILD',
            tf=self.tf,
            side=side,
            quantity=quantity,
            limit_price=limit_price,
            TP=TP,
            SL=SL,
            confidence=self.confidence,
            executionDepends=executionDepends,
            periodUnfilled=periodUnfilled,
            periodTillTp=periodTillTp,
        )
        self.CHILDREN[childInsight.INSIGHT_ID] = childInsight
        return childInsight

    def updateState(self, state: InsightState, message: str = None):
        print(
            f"Updated Insight State: {self.state:^10} -> {state:^10}: {self.symbol:^8} : {self.strategyType} :", message)
        self.state = state
        self.updatedAt = datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time
        if self.state == InsightState.FILLED:
            if self._cancelling:
                self._cancelling = False
            self.filledAt = self.updatedAt
        if self.state == InsightState.CLOSED:
            self.closedAt = self.updatedAt
            # Print the P/L of the trade
            print(self.logPnL())

        return self

    def update_quantity(self, quantity: float):
        if not  self.uses_contract_size:
            old_quantity = self.quantity
            self.quantity = quantity
            if self.checkValidQuantity():
                print(f"Updated quantity: {old_quantity} -> {self.quantity}")
                return True
            else:
                self.quantity = old_quantity
                return False
        else:
            return self.update_contracts(quantity)

    def update_contracts(self, contracts: float):
        if self.uses_contract_size:
            old_contracts = self.contracts
            self.contracts = contracts
            if self.checkValidQuantity():
                print(f"Updated contracts: {
                      old_contracts} -> {self.contracts}")
                self.quantity = self.contracts * self.ASSET["contract_size"]
                return True
            else:
                self.contracts = old_contracts
                return False
        else:
            print("Asset has no contract size")
            return False

    def update_limit_price(self, price: float, updateToLimit: bool = False):
        # check if price is the same as the limit price
        if price == self.limit_price:
            print(
                f"Limit price is the same as the current limit price: {price} == {self.limit_price}")
            return True
        # check if the insight is already filled
        if self.state == InsightState.FILLED:
            print(
                f"Insight is already filled: {self.symbol} - {self.side} - {self.quantity} @ {self.limit_price}")
            return False
        # check if the price is within the take profit and stop loss
        if not self.checkValidEntryInsight(price):
            return False

        self.limit_price = price

        self.updatedAt = datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time

        if updateToLimit:
            self.type = IOrderType.LIMIT

        # check if the insight is already executed
        if self.state == InsightState.EXECUTED:

            # TODO: update the broker order if the insight is already placed
            # TODO: Need to check if the order is already placed
            #  self.BROKER ....
            pass
        return True

    # TODO: def update_take_profit(self, price: float):
    #     if self.state == InsightState.FILLED:
    #         pass
    # TODO: def update_stop_loss(self, price: float):
    #     if self.state == InsightState.FILLED:
    #         pass
    def update_market_changed(self, marketChanged: bool, shouldCloseOrCancel: bool = False):
        self.marketChanged = marketChanged
        self.updatedAt = datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time
        if shouldCloseOrCancel:
            if self.state == InsightState.FILLED:
                if self.close():
                    return True
            else:
                if self.cancel():
                    return True
            # Failed to close or cancel the insight
            return False
        return True

    def validate(self) -> tuple[bool, Optional[str]]:
        """Validate the insight before submitting it to the broker."""
        cause = None
        # Skip the validation for manual and test insights
        if (self.strategyType == StrategyTypes.TEST or self.strategyType == StrategyTypes.MANUAL):
            return (True, None)

        if not self.checkValidEntryInsight():
            cause = 'Invalid Entry Insight'
            return (False, cause)
        if not self.checkIfCanShort(shouldUpdateState=False):
            cause = 'Short not allowed'
            return (False, cause)
        if self.state == InsightState.FILLED:
            if self.hasExhaustedTTL(shouldUpdateState=False):
                cause = 'Filled Time To Live expired'
                return (False, cause)
        else:
            if self.hasExpired(shouldUpdateState=False):
                cause = 'Unfilled Time To Live expired'
                return (False, cause)
        if not self.checkValidQuantity(shouldUpdateState=False):
            cause = 'Invalid Quantity'
            return (False, cause)

        if self.marketChanged:
            cause = 'Market has changed'
            return (False, cause)

        return (True, cause)

    def checkValidQuantity(self, shouldUpdateState: bool = False):
        # Check if quantity is invalid
        if self.uses_contract_size and self.contracts:
            if self.contracts == None or self.ASSET['min_order_size'] > self.contracts or isnan(self.contracts):
                if shouldUpdateState:
                    self.updateState(
                        InsightState.REJECTED, 'Invalid Contracts')
                return False
        elif self.quantity == None or self.ASSET['min_order_size'] > self.quantity or isnan(self.quantity):
            if shouldUpdateState:
                self.updateState(
                    InsightState.REJECTED, 'Invalid Quantity')
            return False

        return True

    def checkIfCanShort(self, shouldUpdateState: bool = False):
        """Check if the asset is shortable. If the asset is not shortable, the insight will be rejected."""
        # Skip the check if going long
        if self.side == IOrderSide.BUY:
            return True
        # Check if the asset is shortable
        if ((self.side == IOrderSide.SELL) and self.ASSET['shortable']):
            return True

        if shouldUpdateState:
            self.updateState(
                InsightState.REJECTED, f"Short not allowed")
        return False

    def checkValidEntryInsight(self, limit_price: float = None):
        """Check if the insight is valid. limitprice needs to be beween the take profit and stop loss."""
        if (self.strategyType == StrategyTypes.TEST or self.strategyType == StrategyTypes.MANUAL):
            return True  # skip the check for manual and test insights

        limit_price = limit_price if limit_price != None else self.limit_price
        # FIXME: use the broker to get the latest price and to check if the market order would be within range of the bracket order
        if limit_price == None:
            print("WARNING: limit price is not set, Using Market Order")
            return True

        if self.SL:
            if (limit_price < self.SL and self.side == IOrderSide.BUY) or (limit_price > self.SL and self.side == IOrderSide.SELL):
                print("invalid entry insight: limit price is below the stop loss")
                return False
        if self.TP:
            if len(self.TP) == 1:
                if (limit_price > self.TP[0] and self.side == IOrderSide.BUY) or (limit_price < self.TP[0] and self.side == IOrderSide.SELL):
                    print("invalid entry insight: limit price is below the take profit")
                    return False
            else:
                if self.side == IOrderSide.BUY:
                    self.TP.sort()
                else:
                    self.TP.sort(reverse=True)
                for tp in self.TP:
                    if (limit_price > tp and self.side == IOrderSide.BUY) or (limit_price < tp and self.side == IOrderSide.SELL):
                        print(
                            "invalid entry insight: limit price is above the take profit")
                        return False
        else:
            print("invalid entry insight: quantity is not set")
            return False

        return True

    def hasExpired(self, shouldUpdateState: bool = False) -> bool | None:
        """Check if the insight has expired. If the insight is not filled within the time to live period, it will be canceled. If the insight is filled, it will be closed after the time to live till take profit has expired."""
        if self.periodUnfilled == None:
            return False

        if self.state == InsightState.FILLED:
            # Insight has already been filled or canceled
            return None

        expireAt = self.tf.add_time_increment(
            self.createAt, self.periodUnfilled)
        hasExpired = expireAt < (datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time)

        if (self.state == InsightState.NEW or self.state == InsightState.EXECUTED) and hasExpired and shouldUpdateState:
            if self.state == InsightState.EXECUTED:
                if self.cancel():
                    pass
                    # self.updateState(InsightState.CANCELED,
                    #                  'Unfilled Time To Live expired')
            else:
                self.updateState(InsightState.REJECTED,
                                 'Expired before execution')

        return hasExpired

    def hasExhaustedTTL(self, shouldUpdateState: bool = False):
        if self.periodTillTp == None:
            return False

        if self.state != InsightState.FILLED:
            # Insight has not been filled yet
            return None

        expireAt = self.tf.add_time_increment(
            self.filledAt, self.periodTillTp)
        hasExpired = expireAt < (datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time)
        if (self.state == InsightState.FILLED) and hasExpired and shouldUpdateState:
            if self.close():
                # Close state switch should handled by the strategy
                # self.updateState(InsightState.CLOSED,
                #                  'Filled Time To Live expired')
                pass

        return hasExpired

    def updateOrderID(self, order_id: str):
        self.order_id = order_id
        self.updatedAt = datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time
        return self

    def updateCloseOrderID(self, close_order_id: str):
        self.close_order_id = close_order_id
        self.updatedAt = datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time
        return self

    def updateLegs(self, legs: IOrderLegs):
        """Update the order legs for the insight."""
        self.legs = legs
        self.updatedAt = datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time
        return self

    def updateTakeProfitLegs(self, leg: IOrderLeg):
        # TODO: Check if there is a take profit order leg already set if so update the limit price

        self.legs['take_profit'] = leg
        self.TP[-1] = leg['limit_price']
        self.updatedAt = datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time
        return self

    @property
    def takeProfitOrderLeg(self):
        return self.legs.get('take_profit', None)

    def cancelTakeProfitLeg(self):
        if self.takeProfitOrderLeg:
            if self.cancel_order_by_id(self.takeProfitOrderLeg['order_id']) or (self.takeProfitOrderLeg['order_id'] == None or self.takeProfitOrderLeg['order_id'] == ""):

                self.legs['take_profit'] = None
                self.updatedAt = datetime.now(
                ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time
                return True
        return False

    def updateStopLossLegs(self, leg: IOrderLeg):
        # TODO: Check if there is a stop loss order leg already set if so update the limit price
        self.legs['stop_loss'] = leg
        self.SL = leg['limit_price']
        self.updatedAt = datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time
        return self

    @property
    def stopLossOrderLeg(self):
        return self.legs.get('stop_loss', None)

    def cancelStopLossLeg(self):
        if self.stopLossOrderLeg:
            if self.cancel_order_by_id(self.stopLossOrderLeg['order_id'] or (self.stopLossOrderLeg['order_id'] == None or self.stopLossOrderLeg['order_id'] == "")):
                self.legs['stop_loss'] = None
                self.updatedAt = datetime.now(
                ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time
                return True
        return False

    def updateTrailingStopLegs(self, leg: IOrderLeg):
        self.legs['trailing_stop'] = leg
        self.updatedAt = datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time
        return self

    @property
    def trailingStopOrderLeg(self):
        return self.legs.get('trailing_stop')

    def positionFilled(self, price: float, qty: float, order_id: str = None):
        if self.order_id == None:
            self.updateOrderID(order_id)

        assert self.order_id == order_id, f"Order ID: {
            order_id} does not match the insight's order ID: {self.order_id}"
        # TODO: Some times for market orders the price is not set so we will need to run the TP and SL calculations based on the current filled price.
        # We may want to give the traders the option to set a new TP and SL based on the current price. or just move the TP and SL based on the current price to match the minumem RRR.
        self.limit_price = price
        if self.uses_contract_size:
            self.update_contracts(qty)
        else:
            self.update_quantity(qty)
        self.partialFilled(qty)
        self.updateState(InsightState.FILLED, f"Trade Filled: {
                         self.symbol} - {self.side} - {self.quantity} @ {self.limit_price}")

        return self

    def partialFilled(self, qty: float):
        if self.uses_contract_size:
            qty = qty * self.ASSET["contract_size"]
        if self._partial_filled_quantity == None:
            self._partial_filled_quantity = 0
        if qty <= self.quantity:
            self._partial_filled_quantity = qty
            return self
        else:
            print(
                f"Partial Filled Quantity: {qty} is greater than the insight's position size: {self.quantity}")
            return

    def positionClosed(self, price: float, close_order_id: str, qty: float = None):
        if self.close_order_id == None:
            self.updateCloseOrderID(close_order_id)
        assert ((self.close_order_id == close_order_id) or (self.order_id == close_order_id)), f"Close Order ID: {
            close_order_id} does not match the insight's close order ID: {self.close_order_id}"
        if qty != None:
            self.update_quantity(qty)
        # Handle multiple Take Profit levels
        if self.close_price != None:
            self.close_price += price
            self.close_price /= 2
        else:
            self.close_price = price

        self.updateState(InsightState.CLOSED)

        # Close all of the child insights
        if self.PARANT == None and len(self.CHILDREN) > 0:
            for child in self.CHILDREN:
                childInsight = self.CHILDREN[child]
                if childInsight.state == InsightState.FILLED:
                    childInsight.close()
                elif childInsight.state == InsightState.EXECUTED:
                    childInsight.cancel()
                else:  # childInsight.state == InsightState.NEW, InsightState.REJECTED, InsightState.CANCELED
                    childInsight.updateState(
                        InsightState.CANCELED, 'Parent Insight Closed')
        return self

    def getPL(self):
        assert self.close_price != None, 'Close price is not set'
        partialPL = 0
        if len(self.partial_closes) > 0:
            for partial in self.partial_closes:
                partialPL += partial.getPL()
        
        if self.side == IOrderSide.BUY:
            return round(((self.close_price - self.limit_price) * self.quantity) - partialPL, 2)
        else:
            return round(((self.limit_price - self.close_price) * self.quantity) - partialPL, 2)

    def getPnLRatio(self):
        if self.TP and self.SL and self.limit_price != None:
            return round((abs(self.TP[-1] - self.limit_price)) / (abs(self.limit_price - self.SL)), 2)
        else:
            return 0

    def logPnL(self):
        PL = self.getPL()
        message = f"Trade Closed {"✅" if PL > 0 else "❌"}: {self.symbol} - {self.side} - {
            self.quantity} @ {self.close_price} - P/L: {PL} - UDA: {self.updatedAt}"
        return message

    def set_mode(self, broker: get_BaseBroker, asset: IAsset, mode: IStrategyMode = IStrategyMode.LIVE):
        self.MODE = mode
        self.BROKER = broker
        self.ASSET = asset
        # update the created at to the current time in the simulation
        self.createAt = datetime.now(
        ) if self.MODE == IStrategyMode.LIVE else self.BROKER.get_current_time
        self.updatedAt = self.createAt

        if self.checkValidEntryInsight():
            print(f"Created Insight: {self}")
        else:
            self.updateState(InsightState.REJECTED, 'Invalid Entry Insight')

        return self

    @property
    def uses_contract_size(self):
        if self.ASSET == None:
            return False
        return self.ASSET.get("contract_size", None) != None

    @property
    def dataclass(self):
        return IInsight(self)


@dataclass
class IInsight:
    symbol: str
    insight_id: str
    parent: Optional[str]
    children: dict[UUID, Self]
    state: str
    side: str
    limit_price: float
    type: str
    take_profit: Optional[float]
    stop_loss: Optional[float]
    useContractSize: bool
    quantity: float
    contracts: Optional[float]
    strategy: Optional[str]
    execution_dependency: str
    order_id: str
    confidence: float
    RRR: float
    TTL_unfilled: float
    TTL_filled: float
    close_order_id: Optional[str]
    close_price: Optional[float]
    created_at: datetime
    updated_at: datetime
    filled_at: Optional[datetime]
    closed_at: Optional[datetime]
    legs: IOrderLegs
    partial_filled_quantity: float
    partial_closes: list[PartialCloseResult]
    closing: bool
    cancelling: bool
    marketChanged: bool

    def __init__(self, insight: Insight):
        self.symbol = insight.symbol
        self.insight_id = str(insight.INSIGHT_ID)
        self.parent = str(
            insight.PARANT) if insight.PARANT is not None else None
        self.children = {key: IInsight(child)
                         for key, child in insight.CHILDREN.items()}
        self.state = insight.state.value
        self.side = insight.side.value
        self.limit_price = insight.limit_price
        self.type = str(insight.type.value)
        self.take_profit = insight.TP
        self.stop_loss = insight.SL
        self.useContractSize = insight.uses_contract_size
        self.quantity = insight.quantity
        self.contracts = insight.contracts if insight.uses_contract_size else None
        
        self.strategy = str(insight.strategyType) if type(
            insight.strategyType) == str else insight.strategyType.value if insight.strategyType != None else None
        self.execution_dependency = [
            str(executionDepends.value) for executionDepends in insight.executionDepends]
        self.order_id = str(insight.order_id)
        self.confidence = insight.confidence
        self.RRR = insight.getPnLRatio() or 0
        self.TTL_unfilled = insight.periodUnfilled
        self.TTL_filled = insight.periodTillTp
        self.close_order_id = str(insight.close_order_id)
        self.close_price = insight.close_price
        self.created_at = insight.createAt
        self.updated_at = insight.updatedAt
        self.filled_at = insight.filledAt
        self.closed_at = insight.closedAt
        legs = {}
        if insight.takeProfitOrderLeg:
            legs['take_profit'] = {
                'order_id': str(insight.takeProfitOrderLeg['order_id']),
                'limit_price': insight.takeProfitOrderLeg['limit_price'],
                'filled_price': insight.takeProfitOrderLeg.get('filled_price',None),
                'type': str(insight.takeProfitOrderLeg.get('type', None)),
                'status': str(insight.takeProfitOrderLeg['status']),
                'order_class': str(insight.takeProfitOrderLeg.get('order_class', None)),
                'created_at': insight.takeProfitOrderLeg.get('created_at', None),
                'updated_at': insight.takeProfitOrderLeg.get('updated_at', None),
                'submitted_at': insight.takeProfitOrderLeg.get('submitted_at', None),
                'filled_at': insight.takeProfitOrderLeg.get('filled_at', None),
            }
        if insight.stopLossOrderLeg:
            legs['stop_loss'] = {
                'order_id': str(insight.stopLossOrderLeg['order_id']),
                'limit_price': insight.stopLossOrderLeg['limit_price'],
                'filled_price': insight.stopLossOrderLeg['filled_price'],
                'type': str(insight.stopLossOrderLeg.get('type')),
                'status': str(insight.stopLossOrderLeg['status']),
                'order_class': str(insight.stopLossOrderLeg.get('order_class', None)),
                'created_at': insight.stopLossOrderLeg.get('created_at', None),
                'updated_at': insight.stopLossOrderLeg.get('updated_at', None),
                'submitted_at': insight.stopLossOrderLeg.get('submitted_at', None),
                'filled_at': insight.stopLossOrderLeg.get('filled_at', None),
            }

        self.legs = legs if legs else None
        self.partial_filled_quantity = insight._partial_filled_quantity
        partialCloses = [{"order_id": str(partialclose.order_id),
                          "side": str(partialclose.side),
                          "quantity": partialclose.quantity,
                          "entry_price": partialclose.entry_price,
                          "filled_price": partialclose.filled_price,
                          } for partialclose in insight.partial_closes]
        self.partial_closes = partialCloses
        self.closing = insight._closing
        self.cancelling = insight._cancelling
        self.marketChanged = insight.marketChanged
        





from ..base_executor import BaseExecutor
from ...insight import InsightState
from ....broker.interfaces import IOrderSide
from ....strategy.interfaces import IStrategyMode
from ....utils.tools import dynamic_round


class BasicTakeProfitExecutor(BaseExecutor):
    """
    ### Executor for Basic Take Profit
    This executor is used to take profit on insights that have reached the take profit price.

    :param strategy (BaseStrategy): The strategy instance
    
    Author:
        @isaac-diaby
    """

    def __init__(self, strategy):
        super().__init__(strategy, InsightState.FILLED, "1.0")

    def run(self, insight):
        # check if the insight already has a take profit order leg
        if insight.takeProfitOrderLeg: #  and self.STRATEGY.MODE != IStrategyMode.BACKTEST:
            return self.returnResults(True, True, "Insight already has a take profit order")

        # Check if the insight has reached the take profit price
        if insight.TP == None:
            return self.returnResults(True, True, "Insight does not have take profit level set.")
        try:
            # check if the insight has not already been closed 
            if insight._closing:
                return self.returnResults(False, True, "Insight is being closed.")
            # Check if price broke the first Take Profit level
            latestBar = self.get_latest_bar(insight.symbol)
            latestQuote = self.get_latest_quote(insight)
            shouldClose = False

            currentTP = insight.TP[0]
            match insight.side:
                case IOrderSide.BUY:
                    if (latestBar.high > currentTP) or (latestQuote["bid"] > currentTP):
                        shouldClose = True
                case IOrderSide.SELL:
                    if (latestBar.low < currentTP) or (latestQuote["ask"] < currentTP):
                        shouldClose = True
            if shouldClose:
                if len(insight.TP) > 1:
                    quantityToClose = dynamic_round(
                        insight.quantity/2)

                    if self.STRATEGY.insights[insight.INSIGHT_ID].close(quantity=quantityToClose):
                        return self.returnResults(False, True, f"Price broke the take profit level: {insight.symbol} : {currentTP}. Closing half position.")
                else:
                    # Close the position if the last TP level is reached
                    if self.STRATEGY.insights[insight.INSIGHT_ID].close():
                        return self.returnResults(False, True, f"Price broke the take profit level: {insight.symbol} : {currentTP}. Closing position.")
            return self.returnResults(True, True, f"Take profit price has not been reached yet: {insight.symbol} : {currentTP}")
        except Exception as e:
            return self.returnResults(False, False, f"Error taking profit: {e}")



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import List, Literal, NotRequired, Optional, TypedDict, Union
from uuid import UUID

@dataclass
class ISupportedBrokerFeatures:
    """Supported Broker Features."""
    barDataStreaming: bool = True
    tradeEventStreaming: bool = True
    featuredBarDataStreaming: bool = True
    submitOrder: bool = True
    maxOrderValue: float = None
    cancelOrder: bool = True
    closePosition: bool = True
    getAccount: bool = True
    getPosition: bool = True
    getPositions: bool = True
    getHistory: bool = True
    getQuote: bool = True
    getTickerInfo: bool = True
    leverage: bool = True
    shorting: bool = True
    margin: bool = True
    bracketOrders: bool = True
    trailingStop: bool = True


class ISupportedBrokers(Enum):
    """Supported Brokers."""
    ALPACA = 'AlpacaBroker'
    CCXT = 'CCXTBroker'
    MT5 = 'MetaTrader5Broker'
    PAPER = 'PapeVrokerr'
    BASE = 'BaseBroker'


class ITradeUpdateEvent(Enum):
    """Trade Update Event."""
    ACCEPTED = 'accepted'
    NEW = 'new'
    PENDING_NEW = 'pending_new'
    PARTIAL_FILLED = 'partial_filled'
    FILLED = 'fill'
    CANCELED = 'canceled'
    REJECTED = 'rejected'
    EXPIRED = 'expired'
    CLOSED = 'closed'
    REPLACED = 'replaced'


class ITimeInForce(Enum):
    DAY = 'day'
    GTC = 'gtc'
    OPG = 'opg'
    IOC = 'ioc'
    FOK = 'fok'


class IOrderClass(Enum):
    # Simple Order - with no legs
    SIMPLE = 'simple'
    # Bracket Order - (otooco) # One Triggers One Cancels Other (Take Profit and Stop Loss)
    BRACKET = 'bracket'
    # One Cancels Other (Take Profit and Stop Loss) - on active order
    OCO = 'oco'
    # One Triggers Other (Take Profit or Stop Loss) - on active order
    OTO = 'oto'
    TRO = 'tro'  # Trailing Stop Order


class IOrderType(Enum):
    """Order Type."""
    MARKET = 'Market'
    LIMIT = 'Limit'
    STOP = 'Stop'
    STOP_LIMIT = 'Stop_limit'
    TRAILING_STOP = 'Trailing_stop'


class IOrderSide(Enum):
    """Order Side."""
    BUY = 'Long'
    SELL = 'Short'


class IOrderRequest(TypedDict):
    symbol: str
    qty: float
    side: IOrderSide
    type: IOrderType
    time_in_force: ITimeInForce
    limit_price: float
    order_class: IOrderClass
    take_profit: float
    stop_loss: float
    trail_price: float


class IAsset(TypedDict):
    id: str
    name: str
    asset_type: Literal['stock', 'crypto', 'forex']
    exchange: str
    symbol: str
    status: Literal['active', 'inactive']
    tradable: bool
    marginable: bool
    shortable: bool
    fractionable: bool
    min_order_size: float
    max_order_size: NotRequired[float]
    min_price_increment: float
    price_base: NotRequired[int]
    contract_size: NotRequired[int] # For futures and options contracts only (in the asset's base currency). else None

@dataclass
class IAccount():
    account_id: str
    equity: float
    cash: float
    currency: str
    buying_power: float
    shorting_enabled: bool
    leverage: float

class IPosition(TypedDict):
    asset: IAsset
    avg_entry_price: float
    qty: float
    side: IOrderSide
    market_value: float
    cost_basis: float
    current_price: float
    unrealized_pl: float

class IOrderLeg(TypedDict):
    order_id: str
    limit_price: float
    filled_price: Optional[float]
    type: IOrderType
    status: ITradeUpdateEvent
    order_class: IOrderClass
    created_at: datetime
    updated_at: datetime
    submitted_at: datetime
    filled_at: Optional[datetime]

class IOrderLegs(TypedDict):
    take_profit: Optional[IOrderLeg]
    stop_loss: Optional[IOrderLeg]
    trailing_stop: Optional[IOrderLeg]

class IOrder(TypedDict):
    order_id: Union[str, UUID]
    asset: IAsset
    limit_price: float
    filled_price: Optional[float]
    stop_price: Optional[float]
    qty: float
    filled_qty: float
    side: IOrderSide
    type: IOrderType
    time_in_force: ITimeInForce
    status: ITradeUpdateEvent
    order_class: IOrderClass
    created_at: datetime
    updated_at: datetime
    submitted_at: datetime #Timestamp when the order was submitted.
    filled_at: Optional[datetime] #Timestamp when the order was filled.
    legs: Optional[IOrderLegs]

class IQuote(TypedDict):
    symbol: str
    bid: float
    ask: float
    bid_size: float
    ask_size: float
    volume: float
    timestamp: datetime


class IAccountState(TypedDict):
    account: IAccount
    positions: dict[str, IPosition]
    orders: List[IOrder]

class ITradeUpdate():
    def __init__(self, order: IOrder, event: ITradeUpdateEvent):
        self.event = event
        self.order = order

    def __str__(self):
        return f'{self.event} - {self.order["asset"]["symbol"]} - {self.order["qty"]} - {self.order["side"]} - {self.order["updated_at"]}'



import datetime
from typing import Awaitable, Callable, List, Literal, Union
from concurrent.futures import ThreadPoolExecutor
import asyncio
import os
import pandas as pd
import json
import numpy as np

import alpaca
from alpaca.trading.client import TradingClient
from alpaca.data.historical import StockHistoricalDataClient, CryptoHistoricalDataClient
from alpaca.data.live import StockDataStream, CryptoDataStream
from alpaca.data.requests import StockBarsRequest, CryptoBarsRequest, StockLatestQuoteRequest, CryptoLatestQuoteRequest
from alpaca.data.timeframe import TimeFrame
from alpaca.data.enums import DataFeed, CryptoFeed
from alpaca.common.enums import BaseURL
from alpaca.trading.models import Position, Order, TradeUpdate
from alpaca.data.models import Bar, Quote
from alpaca.trading.requests import MarketOrderRequest, LimitOrderRequest, StopLimitOrderRequest, ClosePositionRequest, OrderSide, OrderType, OrderClass, TimeInForce, TakeProfitRequest, StopLossRequest
from alpaca.trading.stream import TradingStream
# from alpaca.trading.enums import AssetClass

from .base_broker import BaseBroker
from .interfaces import ISupportedBrokerFeatures, ISupportedBrokers, IAsset, IAccount, IOrder, IPosition, IOrderSide, IOrderType, ITradeUpdateEvent, IQuote, IOrderLegs, IOrderLeg
from ..utils.timeframe import ITimeFrame as tf
from ..insight.insight import Insight


class AlpacaBroker(BaseBroker):
    trading_client: TradingClient = None
    trade_stream_client: TradingStream = None
    stock_client: StockHistoricalDataClient = None
    stock_stream_client: StockDataStream = None
    crypto_client: CryptoHistoricalDataClient = None
    crypto_stream_client: CryptoDataStream = None

    def __init__(self, paper: bool, feed: DataFeed = DataFeed.IEX):
        super().__init__(ISupportedBrokers.ALPACA, paper, feed)

        assert os.getenv('ALPACA_API_KEY'), 'ALPACA_API_KEY not found'
        assert os.getenv('ALPACA_SECRET_KEY'), 'ALPACA_SECRET_KEY not found'

        self.DataFeed = DataFeed.IEX if self.PAPER else DataFeed.SIP

        self.trading_stream_client = TradingStream(
            os.getenv('ALPACA_API_KEY'),  os.getenv('ALPACA_SECRET_KEY'), paper=paper)
        self.trading_client = TradingClient(
            os.getenv('ALPACA_API_KEY'),  os.getenv('ALPACA_SECRET_KEY'), paper=paper)
        self.stock_client = StockHistoricalDataClient(
            os.getenv('ALPACA_API_KEY'),  os.getenv('ALPACA_SECRET_KEY'))
        self.crypto_client = CryptoHistoricalDataClient(
            os.getenv('ALPACA_API_KEY'),  os.getenv('ALPACA_SECRET_KEY'))
        self.stock_stream_client = StockDataStream(
            os.getenv('ALPACA_API_KEY'),  os.getenv('ALPACA_SECRET_KEY'))
        self.crypto_stream_client = CryptoDataStream(
            os.getenv('ALPACA_API_KEY'),  os.getenv('ALPACA_SECRET_KEY'), feed=CryptoFeed.US, url_override=BaseURL.MARKET_DATA_STREAM + "/v1beta3/crypto/" + CryptoFeed.US)

        self.supportedFeatures = ISupportedBrokerFeatures(
            barDataStreaming=True, trailingStop=False, maxOrderValue=200_000.00)

    def get_history(self, asset: IAsset, start: datetime, end: datetime, resolution: tf):
        # Convert to TimeFrame from OlympusTrader from alpaca
        super().get_history(asset, start, end, resolution)

        timeframe = TimeFrame(resolution.amount, resolution.unit)
        data = None
        try:
            if (asset['asset_type'] == 'stock'):
                # assert isinstance(
                #     feed, DataFeed), 'DataFeed must be of type DataFeed'
                data = self.stock_client.get_stock_bars(
                    StockBarsRequest(
                        symbol_or_symbols=asset['symbol'],
                        timeframe=timeframe,
                        start=start,
                        end=end,
                        feed=self.DataFeed
                    )).df
            elif (asset['asset_type'] == 'crypto'):
                # assert isinstance(
                #     feed, CryptoFeed), 'DataFeed must be of type CryptoFeed'
                data = self.crypto_client.get_crypto_bars(CryptoBarsRequest(
                    symbol_or_symbols=asset['symbol'],
                    timeframe=timeframe,
                    start=start,
                    end=end
                )).df
            else:
                assert False, 'Get History: IAsset type must be of type stock or crypto'

            if data.empty:
                return None
            # format Data Frame open, high, low, close, volume
            # 'timestamp'
            data = data[['open', 'high', 'low', 'close', 'volume']]
            # print(data)
            return self.format_on_bar(data)
        except:
            return None

    def get_ticker_info(self, symbol):
        try:
            if symbol in self.TICKER_INFO:
                return self.TICKER_INFO[symbol]

            tickerInfo = self.trading_client.get_asset(symbol)

            assert tickerInfo, f'Asset {symbol} not found'

            tickerAsset: IAsset = IAsset(
                id=tickerInfo.id,
                name=tickerInfo.name,
                asset_type='stock' if tickerInfo.asset_class == 'us_equity' else 'crypto',
                exchange=tickerInfo.exchange,
                symbol=tickerInfo.symbol,
                status=tickerInfo.status,
                tradable=tickerInfo.tradable,
                marginable=tickerInfo.marginable,
                shortable=tickerInfo.shortable,
                fractionable=tickerInfo.fractionable,
                min_order_size=tickerInfo.min_order_size if tickerInfo.min_order_size else 0.1,
                min_price_increment=tickerInfo.price_increment if tickerInfo.price_increment else 0.01,
            )
            self.TICKER_INFO[symbol] = tickerAsset
            return tickerAsset
        except alpaca.common.exceptions.APIError as e:
            print("Error: No asset for", symbol)
            return None

    def get_account(self):
        res = self.trading_client.get_account()
        res.non_marginable_buying_power
        account: IAccount = IAccount(account_id=str(res.id), equity=float(res.equity), cash=float(res.cash), currency=res.currency,
                                     buying_power=float(res.buying_power), leverage=float(res.multiplier), shorting_enabled=res.shorting_enabled)
        return account

    def get_position(self, symbol):
        res = self.trading_client.get_position(symbol)
        position: IPosition = self.format_position(position)
        return position

    def get_positions(self):
        res: List[Position] = self.trading_client.get_all_positions()

        positions: dict[str, IPosition] = {}
        for position in res:
            positions[position.symbol] = self.format_position(position)
        return positions

    def format_position(self, position: Position) -> IPosition:
        return IPosition(
            asset=self.get_ticker_info(position.symbol),
            avg_entry_price=float(position.avg_entry_price),
            qty=float(position.qty),
            side=position.side.value,
            market_value=float(position.market_value),
            cost_basis=float(position.cost_basis),
            current_price=float(position.current_price),
            unrealized_pl=float(position.unrealized_pl)
        )

    def get_orders(self):
        res = self.trading_client.get_orders()
        orders: dict[str, IOrder] = {}
        if not res:
            return None
        for order in res:
            orders[order.id] = self.format_order(order)
        return orders

    def get_order(self, order_id):
        return self.format_order(self.trading_client.get_order_by_id(order_id))

    def format_order(self, order: Order) -> IOrder:
        side = None
        match order.side:
            case OrderSide.BUY:
                side = IOrderSide.BUY
            case OrderSide.SELL:
                side = IOrderSide.SELL
            case _:
                side = "unknown"
        #  TODO: add support for order legs
        legs = IOrderLegs()
        if order.legs:
            for leg in order.legs:
                if leg.order_type == OrderType.LIMIT:
                    legs['take_profit'] = IOrderLeg(
                        order_id=leg.id,
                        limit_price=float(leg.limit_price),
                        filled_price=float(
                            leg.filled_avg_price) if leg.filled_avg_price else None,
                        type=IOrderType.LIMIT,
                        status=leg.status.value,
                        order_class=leg.order_class.value,
                        created_at=leg.created_at,
                        updated_at=leg.updated_at,
                        submitted_at=leg.submitted_at,
                        filled_at=leg.filled_at
                    )
                elif leg.order_type == OrderType.STOP:
                    legs['stop_loss'] = IOrderLeg(
                        order_id=leg.id,
                        limit_price=float(leg.stop_price),
                        filled_price=float(
                            leg.filled_avg_price) if leg.filled_avg_price else None,
                        type=IOrderType.STOP,
                        status=leg.status.value,
                        order_class=leg.order_class.value,
                        created_at=leg.created_at,
                        updated_at=leg.updated_at,
                        submitted_at=leg.submitted_at,
                        filled_at=leg.filled_at
                    )
                else:
                    continue

        order = IOrder(
            order_id=order.id,
            asset=self.get_ticker_info(order.symbol),
            filled_price=float(
                order.filled_avg_price) if order.filled_avg_price else None,
            limit_price=float(
                order.limit_price) if order.limit_price else None,
            stop_price=float(order.stop_price) if order.stop_price else None,
            qty=float(order.qty),
            filled_qty=float(order.filled_qty),
            side=side,
            type=order.type.value,
            order_class=order.order_class.value,
            time_in_force=order.time_in_force.value,
            status=order.status.value,
            created_at=order.created_at,
            updated_at=order.updated_at,
            submitted_at=order.submitted_at,
            filled_at=order.filled_at,
            legs=legs
        )

        return order

    def get_latest_quote(self, asset: IAsset):
        if asset['asset_type'] == 'crypto':
            quote = self.crypto_client.get_crypto_latest_quote(
                CryptoLatestQuoteRequest(symbol_or_symbols=asset['symbol']))
        elif asset['asset_type'] == 'stock':
            quote = self.stock_client.get_stock_latest_quote(
                StockLatestQuoteRequest(symbol_or_symbols=asset['symbol'], feed=self.DataFeed))
        return self.format_on_quote(quote[asset['symbol']])

    def execute_insight_order(self, insight: Insight, asset: IAsset) -> IOrder | None:
        # TODO: manage insight order by planing entry and exit orders for a given insight
        # https://alpaca.markets/docs/trading/orders/#bracket-orders
        super().execute_insight_order(insight, asset)
        req = None
        orderRequest = {
            "symbol": insight.symbol,
            "qty": insight.quantity,
            "side": OrderSide.BUY if insight.side == IOrderSide.BUY else OrderSide.SELL,
            "time_in_force": TimeInForce.GTC,
            # OrderClass.BRACKET if insight.TP and insight.SL else OrderClass.SIMPLE,
            "order_class": OrderClass.SIMPLE
            # "take_profit": None if insight.TP == None else TakeProfitRequest(
            #     limit_price=insight.TP[-1]
            # ),
            # "stop_loss": None if insight.SL == None else StopLossRequest(
            #     stop_price=insight.SL,
            #     # stop_price=round(insight.SL-0.01 if insight.side == 'long' else insight.SL+0.01, 2),
            # )
        }
        if insight.TP and insight.SL:
            orderRequest["order_class"] = OrderClass.BRACKET
            orderRequest["take_profit"] = TakeProfitRequest(
                limit_price=insight.TP[-1]
            )
            orderRequest["stop_loss"] = StopLossRequest(
                stop_price=insight.SL,
                # stop_price=round(insight.SL-0.01 if insight.side == 'long' else insight.SL+0.01, 2),
            )
        elif insight.TP:
            orderRequest["order_class"] = OrderClass.OTO
            orderRequest["take_profit"] = TakeProfitRequest(
                limit_price=insight.TP[-1]
            )
        elif insight.SL:
            orderRequest["order_class"] = OrderClass.OTO
            orderRequest["stop_loss"] = StopLossRequest(
                stop_price=insight.SL,
                # stop_price=round(insight.SL-0.01 if insight.side == 'long' else insight.SL+0.01, 2),
            )
        if insight.limit_price:
            orderRequest["limit_price"] = insight.limit_price

        if asset['asset_type'] == 'crypto':
            # "crypto orders not allowed for advanced order_class: otoco or OTO"}
            orderRequest["order_class"] = OrderClass.SIMPLE

        try:
            match insight.type:
                case IOrderType.MARKET:
                    req = MarketOrderRequest(**orderRequest)
                case IOrderType.LIMIT:
                    req = LimitOrderRequest(**orderRequest)
                case IOrderType.STOP:
                    req = StopLimitOrderRequest(**orderRequest)
                case _:
                    print(
                        f"ALPACA: Order Type not supported {insight.type} ")
                    return
            if req:
                order = self.trading_client.submit_order(req)
                return self.format_order(order)

        except alpaca.common.exceptions.APIError as e:
            # print ("ALPACA: Error submitting order", e)
            if e.code == 40310000:
                # '{"available":"0.119784","balance":"0.119784","code":40310000,"message":"insufficient balance for BTC (requested: 0.12, available: 0.119784)","symbol":"USD"}'
                error = BaseException({
                    "code": "insufficient_balance",
                    "data": json.loads(e.args[0])
                })
                # TODO: make a strutured error


                raise error
            raise e

        return None

    def close_all_positions(self):
        print("Closing all positions and orders")
        closeOrders = self.trading_client.close_all_positions(
            cancel_orders=True)
        return closeOrders

    def close_position(self, symbol, qty=None, percent=None):
        try:
            closePosReq = ClosePositionRequest(
                qty=str(qty)) if qty else ClosePositionRequest(percentage=str(percent))
            order = self.trading_client.close_position(symbol, closePosReq)
            # print("Closed position", order)
            return self.format_order(order)
        except alpaca.common.exceptions.APIError as e:
            print("Error closing position", e)
            raise e

    def cancel_order(self, order_id):
        try:
            self.trading_client.cancel_order_by_id(order_id)
            return order_id
        except alpaca.common.exceptions.APIError as e:
            if e.code == 42210000:
                error = BaseException({
                    "code": "already_filled",
                    "data": e
                })
                raise error
            raise e

    def startTradeStream(self, callback: Awaitable):
        super().startTradeStream(callback)
        self.trading_stream_client.subscribe_trade_updates(callback)
        self.trading_stream_client.run()

    async def closeTradeStream(self):
        self.trading_stream_client.stop()
        await self.trading_stream_client.close()

    def streamMarketData(self, callback: Awaitable, assetStreams):
        super().streamMarketData(callback, assetStreams)
        StockStreamCount = 0
        CryptoStreamCount = 0
        barStreamCount = len(
            [asset for asset in assetStreams if asset['type'] == 'bar'])

        pool = ThreadPoolExecutor(max_workers=(
            barStreamCount), thread_name_prefix="MarketDataStream")
        loop = asyncio.new_event_loop()

        for assetStream in assetStreams:

            if assetStream['type'] == 'bar':
                async def alpaca_stream_callback(data):
                    """ Alpaca only generates one bar at a time no matter the timeframe so we need to send the to each feature
                        base_strategy should handle the rest
                    """
                    barData = self.format_on_bar(data)
                    timestamp = barData.index[0][1]
                    # TODO: WE may be able to skip this if python still hass access to the local scope variables assetStream during the loop. 
                    # We may be able to use the assetStream variable directly in the callback instead of looping over it as we are passing by ref to the alpace callback.
                    for asset in assetStreams:
                        if asset['symbol'] == assetStream['symbol'] and asset['time_frame'].is_time_increment(timestamp):
                            # FIXME:Since this is one min candle we need to get the agreg of all of the previous candles if TF is greater than 1 min
                            await callback(barData, timeframe=asset['time_frame'])

                if assetStream.get('feature') != None:
                    # We dont want to add multiple streams for the same asset
                    
                    continue
                if assetStream['asset_type'] == 'stock':
                    StockStreamCount += 1
                    self.stock_stream_client.subscribe_bars(
                        alpaca_stream_callback, assetStream.get('symbol'))
                elif assetStream['asset_type'] == 'crypto':
                    CryptoStreamCount += 1
                    self.crypto_stream_client.subscribe_bars(
                        alpaca_stream_callback, assetStream.get('symbol'))
            else:
                raise NotImplementedError(
                    f"Stream type {assetStream['type']} not supported")
            # TODO: add support for quotes

        if StockStreamCount:
            loop.run_in_executor(pool, self.stock_stream_client.run)
            # self.stock_stream_client.run()
            print(f"Stock Stream Running: {StockStreamCount} streams")
        if CryptoStreamCount:
            loop.run_in_executor(pool, self.crypto_stream_client.run)
            # self.crypto_stream_client.run()
            print(f"Crypto Stream Running: {CryptoStreamCount} streams")
        # TODO: add support for qutoes

    # def streamBar(self, callback: Awaitable, symbol: str, AssetType: Literal['stock', 'crypto'] = 'stock'):
    #     if AssetType == 'stock':
    #         return self.stock_stream_client.subscribe_bars(callback, symbol)
    #     elif AssetType == 'crypto':
    #         return self.crypto_stream_client.subscribe_bars(callback, symbol)

    # def startStream(self, assetType, type):
    #     if assetType == 'stock':
    #         self.stock_stream_client.run()
    #     elif assetType == 'crypto':
    #         self.crypto_stream_client.run()

    async def closeStream(self, assetStreams):
        # TODO: unsubscribe from streams type close stream should be called when strategy is stopped - close WSS connection
        closeStockBarStream = False
        closeCryptoBarStream = False
        for assetStream in assetStreams:
            if assetStream['type'] == 'bar':
                assetType = assetStream['asset_type']
                symbol = assetStream['symbol']
                if assetType == 'stock':
                    self.stock_stream_client.unsubscribe_bars(symbol)
                    if not closeStockBarStream:
                        closeStockBarStream = True
                elif assetType == 'crypto':
                    self.crypto_stream_client.unsubscribe_bars(symbol)
                    if not closeCryptoBarStream:
                        closeCryptoBarStream = True

        if closeStockBarStream:
            self.stock_stream_client.close()
        if closeCryptoBarStream:
            self.crypto_stream_client.stop()

    def format_on_trade_update(self, trade: TradeUpdate):
        event: ITradeUpdateEvent = None
        match trade.event:
            case "fill":
                event = ITradeUpdateEvent.FILLED
            case "partial_fill":
                event = ITradeUpdateEvent.PARTIAL_FILLED
            case "canceled":
                event = ITradeUpdateEvent.CANCELED
            case "rejected":
                event = ITradeUpdateEvent.REJECTED
            case "pending_new":
                event = ITradeUpdateEvent.PENDING_NEW
            case "new":
                event = ITradeUpdateEvent.NEW
            case "expired":
                event = ITradeUpdateEvent.EXPIRED
            case "replaced":
                event = ITradeUpdateEvent.REPLACED
            case "accepted":
                event = ITradeUpdateEvent.ACCEPTED
            case _:
                event = trade.event
        return self.format_order(trade.order), event

    def format_on_bar(self, bar: Union[Bar, pd.DataFrame, pd.Series]):
        if isinstance(bar, pd.DataFrame):
            return bar
        elif isinstance(bar, Bar):
            index = pd.MultiIndex.from_product(
                [[bar.symbol], [bar.timestamp]], names=['symbol', 'date'])
            data = pd.DataFrame(data={
                'open': bar.open,
                'high': bar.high,
                'low': bar.low,
                'close': bar.close,
                'volume': bar.volume,
            }, index=index, columns=['open', 'high', 'low', 'close', 'volume'])
        # data.set_index(['symbol', )], inplace=True)
            return data
        return None

    def format_on_quote(self, quote: Quote):
        data = IQuote(
            symbol=quote.symbol,
            ask=quote.ask_price,
            ask_size=quote.ask_size,
            bid=quote.bid_price,
            bid_size=quote.bid_size,
            volume=(quote.bid_size + quote.ask_size),
            timestamp=quote.timestamp
        )
        return data




from concurrent.futures import ThreadPoolExecutor
import os
from time import sleep
import MetaTrader5 as mt5
from .base_broker import BaseBroker
from .interfaces import IOrderClass, IOrderLeg, IOrderLegs, ISupportedBrokers, ISupportedBrokerFeatures, IAsset, IAccount, IOrder, IPosition, IQuote, ITimeInForce, ITradeUpdate, ITradeUpdateEvent
from .interfaces import IOrderSide, IOrderType
from ..strategy.interfaces import IMarketDataStream
import pandas as pd
from datetime import datetime, timedelta, timezone
from typing import Any, Awaitable, List, Optional, Union
import asyncio
import numpy as np


class Mt5Broker(BaseBroker):
    RUNNING_TRADE_STREAM: bool = False
    """ If The trade stream is running"""
    RUNNING_MARKET_STREAM: bool = False
    """if the market stream is running"""
    _MARKET_STREAMS: dict[IMarketDataStream, asyncio.Future] = {}
    """Market Streams"""

    TF_MAPPING = {
        '1Min': mt5.TIMEFRAME_M1,
        '5Min': mt5.TIMEFRAME_M5,
        '15Min': mt5.TIMEFRAME_M15,
        '30Min': mt5.TIMEFRAME_M30,
        '1Hour': mt5.TIMEFRAME_H1,
        '4Hour': mt5.TIMEFRAME_H4,
        '1Day': mt5.TIMEFRAME_D1,
        '1Week': mt5.TIMEFRAME_W1,
        '1Month': mt5.TIMEFRAME_MN1
    }
    """
    Supported timeframes for MT5 

    TODO: make sure the resolution is in the correct format
    """

    TIMEZONE = timezone.utc
    """MT5 Timezone"""


    def __init__(self, paper: bool, feed=None, timezone = None):
        super().__init__(ISupportedBrokers.MT5, paper, feed)
        print("MetaTrader5 package author: ", mt5.__author__)
        print("MetaTrader5 package version: ", mt5.__version__)

        if not mt5.initialize():
            raise BaseException(
                "initialize() failed, error code =", mt5.last_error())

        assert os.getenv('MT5_LOGIN'), 'MT5_LOGIN not set'
        assert os.getenv('MT5_SECRET_KEY'), 'MT5_SECRET_KEY not set'
        assert os.getenv('MT5_SERVER'), 'MT5_SERVER not set'

        auth = mt5.login(login=int(os.getenv('MT5_LOGIN')), password=os.getenv(
            'MT5_SECRET_KEY'), server=os.getenv('MT5_SERVER'))
        if not auth:
            raise BaseException(
                "Failed to login to MetaTrader 5", mt5.last_error())
        terminal_info=mt5.terminal_info()
        if terminal_info.tradeapi_disabled:
            raise BaseException(
                "Please enable trades from API in MetaTrader 5", mt5.last_error())
        # set time zone

        if timezone:
            self.TIMEZONE = timezone
        else:
            if not self.get_broker_timezone():
                raise BaseException("Failed to get broker's timezone from MetaTrader 5")



        self.supportedFeatures = ISupportedBrokerFeatures(
            barDataStreaming=True, featuredBarDataStreaming=True, trailingStop=False)

    def __del__(self):
        # Shut down connection to the MetaTrader 5
        mt5.shutdown()
    
    def get_broker_timezone(self):
        try:
            symbols=mt5.symbols_get()
            # Get the timezone shift from the first symbol
            # tzdif = pd.Timestamp(symbols[0].time, unit='s') - datetime.now()
            # tzdif = pd.Timestamp(symbols[0].time, unit='s') - pd.Timestamp.now()
            tzdif = pd.Timestamp(symbols[0].time, unit='s', tz=timezone.utc) - pd.Timestamp.now(timezone.utc)
            
            self.TIMEZONE = timezone(tzdif.round(freq='H'))
            # self.TIMEZONE = timezone(np.abs(tzdif).round(freq='H'))
            return True
        except Exception as e:
            print(f"Error: {e}")
            return None


    def get_ticker_info(self, symbol: str) -> Union[IAsset, None]:
        if self.TICKER_INFO.get(symbol):
            return self.TICKER_INFO[symbol]
        try:
            symbol = symbol.replace('/', '')
            tickerInfo = mt5.symbol_info(symbol)
            if not tickerInfo:
                return None
            if not tickerInfo.visible:
                print(symbol, "is not visible, trying to switch on")
                if not mt5.symbol_select(symbol, True):
                    print("Failed to switch on", symbol)
                    None

            # TODO: Store the size of one lot and max contract size
            tradeable = False if tickerInfo.trade_mode == mt5.SYMBOL_TRADE_MODE_DISABLED or (
                tickerInfo.trade_mode == mt5.SYMBOL_TRADE_MODE_CLOSEONLY) else True
            shortable = False if not tradeable or (
                tickerInfo.trade_mode == mt5.SYMBOL_TRADE_MODE_LONGONLY) else True

            fractionable = True if tickerInfo.volume_step < 1 else False

            asset_type = tickerInfo.path.split('\\')[0].lower()

            status = "active" if tickerInfo.visible else "inactive"

            tickerAsset: IAsset = IAsset(
                id=tickerInfo.name,
                name=tickerInfo.description,
                asset_type=asset_type,
                exchange="MT5",
                symbol=symbol,
                status=status,
                tradable=tradeable,
                marginable=True,
                shortable=shortable,
                fractionable=fractionable,
                min_order_size=tickerInfo.volume_min,
                max_order_size=tickerInfo.volume_max,
                min_price_increment=tickerInfo.point,
                price_base=tickerInfo.digits,
                contract_size=tickerInfo.trade_contract_size

            )
            self.TICKER_INFO[symbol] = tickerAsset
            return tickerAsset

        except Exception as e:
            print(f"Error: {e}")
            return None

    def get_account(self) -> IAccount:
        try:
            account_info = mt5.account_info()

            if not account_info:
                return None

            account = IAccount(
                account_id=str(account_info.login),
                equity=account_info.equity,
                cash=account_info.balance,
                currency=account_info.currency,
                buying_power=account_info.margin_free,
                # TODO: Check if this is correct as we are just assuming
                shorting_enabled=account_info.trade_allowed,
                leverage=account_info.leverage
            )
            return account
        except Exception as e:
            print(f"Error: {e}")
            return None

    def get_position(self, symbol) -> IPosition:
        try:
            position = mt5.positions_get(symbol=symbol)
            if not position:
                return None

            position: IPosition = self.format_position(position[0])
            return position
        except Exception as e:
            print(f"Error: {e}")
            return None

    def get_positions(self) -> dict[str, IPosition]:
        try:
            positions = mt5.positions_get()
            if not positions:
                return None

            positions = {position.ticket: self.format_position(
                position) for position in positions}
            # positions = {position.symbol: self.format_position(position) for position in positions}
            return positions

        except Exception as e:
            print(f"Error: {e}")
            return None

    def format_position(self, position: Any) -> IPosition:
        try:
            # Should always be a Trade Posion Class
            return IPosition(
                asset=self.get_ticker_info(position.symbol),
                avg_entry_price=position.price_open,
                qty=position.volume,
                side=IOrderSide.BUY if position.type == 0 else IOrderSide.SELL,
                market_value=position.price_current * position.volume,
                cost_basis=position.price_open * position.volume,
                current_price=position.price_current,
                unrealized_pl=position.profit
            )
        except Exception as e:
            print("Error: {e}")
            return None

    def close_position(self, symbol: str, qty: Optional[float] = None, percent: Optional[float] = None) -> Optional[IOrder]:
        """
        Close a position the param symbol for mt5 is the ticket ID
        """
        super().close_position(symbol, qty, percent)
        try:
            #  Get the position
            positions = self.get_positions()
            if not positions:
                return
            position = positions[symbol]
            if not position:
                return None

            qty_confirmed = None
            # check if we are closing more than we have
            if qty:
                if qty > position['qty']:
                    raise BaseException("Cannot close more than you have")
                else:
                    qty_confirmed = qty
            elif percent:
                qty_confirmed = (position['qty'] * percent)

            # TODO: Will have to test if its position_by or position to close the position  -this looks right for now idk what the difference is rn
            if qty_confirmed != None:
                request = {
                    "action": mt5.TRADE_ACTION_DEAL,
                    "symbol": position['asset'].get('symbol'),
                    "volume": qty_confirmed,
                    "type": mt5.ORDER_TYPE_BUY if position.side == IOrderSide.SELL else mt5.ORDER_TYPE_SELL,
                    "magic": 234777,
                    "comment": "OlympusTrader Close",
                    "type_time": mt5.ORDER_TIME_GTC,
                    "type_filling": mt5.ORDER_FILLING_RETURN,
                    "position": symbol
                    # "position_by": symbol

                }
                result = mt5.order_send(request)
                return self.format_order(result)
            else:
                return None

        except BaseException as e:
            print(e)
            return None

    def close_all_positions(self):
        positions = self.get_positions()
        if not positions:
            return
        for ticket, position in enumerate(positions):
            mt5.Close(symbol=position.symbol, ticket=ticket)

    def get_orders(self) -> Optional[dict[str, IOrder]]:
        try:
            orders: dict[str, IOrder] = {}
        except Exception as e:
            print(f"Error: {e}")
            return None

    def get_order(self, order_id) -> Optional[IOrder]:
        try:
            order = self.format_order(mt5.order_get(ticket=order_id))
            return order
        except Exception as e:
            print(f"Error: {e}")
            return None

    def get_latest_quote(self, asset: IAsset) -> IQuote:
        try:
            quote = mt5.symbol_info_tick(asset.symbol)
            if not quote:
                return None
            return self.format_on_quote(quote, asset.symbol)
        except Exception as e:
            print(f"Error: {e}")
            return None

    def cancel_order(self, order_id: str) -> Optional[str]:
        try:
            request = {
                "action": mt5.TRADE_ACTION_REMOVE,
                "order": order_id,
                "comment": "OlympusTrader Cancel",
            }
            result = mt5.order_send(request)
            if result.retcode != mt5.TRADE_RETCODE_DONE:
                print(f"Order to cancel failed, retcode={
                      result.retcode}, comment={result.comment}")
                return None
            # FIXME: This could also return an error as the position has been filled so cant be cancelled - we will need to check for that.
            return order_id
        except Exception as e:
            print(f"Error: {e}")
            return None

    def get_history(self, asset: IAsset, start, end, resolution) -> pd.DataFrame:
        # TF_MAPPING
        try:
            # exchangeTime = pd.Timestamp(mt5.symbol_info(
            #     asset["symbol"]).time, unit='s', tz=self.TIMEZONE)
            # nowTime = datetime.now(self.TIMEZONE)
            # timeZoneShift = abs(exchangeTime - nowTime)

            # if nowTime > exchangeTime:
            #     start = start - timeZoneShift
            #     end = end - timeZoneShift
            # else:
            #     start = start + timeZoneShift
            #     end = end + timeZoneShift

            rates = mt5.copy_rates_range(
                asset['symbol'], int(resolution), start.astimezone(self.TIMEZONE).replace(tzinfo=timezone.utc), end.astimezone(self.TIMEZONE).replace(tzinfo=timezone.utc))
            # rates = mt5.copy_rates_range(
            #     asset['symbol'], int(resolution), start, end)
            if rates.size < 1:
                return None
            bar = self.format_on_bar(rates, asset['symbol'])

            return bar
        except Exception as e:
            print(f"Error: {e}")
            return None

    def execute_insight_order(self, insight, asset) -> Union[IOrder, None]:
        super().execute_insight_order(insight, asset)
        deviation = 5
        #   Might need to use the below action for limit orders
        # "action": TRADE_ACTION_PENDING

        if asset['contract_size'] > 1:
            volume = insight.contracts
        else:
            volume = insight.quantity
        
        if volume > self.get_ticker_info(insight.symbol).get("max_order_size", np.infty):
            # We could clamp here but user logic would not be aware until after we commit the order.
            print("exceesing max order limit")
            return None 

        si = mt5.symbol_info(asset['symbol'])

        request = {
            "action": mt5.TRADE_ACTION_DEAL,
            "symbol": insight.symbol,
            "volume": volume,
            "type": mt5.ORDER_TYPE_BUY if insight.side == IOrderSide.BUY else mt5.ORDER_TYPE_SELL,
            "deviation": deviation,
            "magic": 234777,
            "comment": f"OlympusTrader Open - {insight.strategyType.value}",
            "type_time": mt5.ORDER_TIME_GTC,
            "type_filling": mt5.ORDER_FILLING_RETURN,
        }
        if insight.type == IOrderType.LIMIT:
            request['type'] = mt5.ORDER_TYPE_BUY_LIMIT if insight.side == IOrderSide.BUY else mt5.ORDER_TYPE_SELL_LIMIT
            # currentPrice = mt5.symbol_info_tick(asset["symbol"]).ask   if insight.side == IOrderSide.BUY else mt5.symbol_info_tick(asset["symbol"]).bid
            # priceDiffInpoints = np.abs(currentPrice - insight.limit_price) * si.point
            # request['price'] =  (currentPrice-priceDiffInpoints  if insight.side == IOrderSide.BUY else currentPrice+priceDiffInpoints) * si.point
            request['price'] = insight.limit_price
            request['action'] = mt5.TRADE_ACTION_PENDING
            # if si.filling_mode == mt5.SYMBOL_TRADE_EXECUTION_MARKET:
            #      request['type_filling'] =  mt5.ORDER_FILLING_BOC
        elif insight.type == IOrderType.MARKET:
            request['type'] = mt5.ORDER_TYPE_BUY if insight.side == IOrderSide.BUY else mt5.ORDER_TYPE_SELL
            request['type_filling'] = mt5.ORDER_FILLING_IOC
            request['price'] = mt5.symbol_info_tick(asset["symbol"]).ask if insight.side == IOrderSide.BUY else mt5.symbol_info_tick(asset["symbol"]).bid
        # set the take profit and stop loss
        if insight.TP:
            # tp_points = round(np.abs(insight.TP[-1] - request['price']) * si.point)
            # request['tp'] = tp_points if insight.side == IOrderSide.BUY else tp_points
            request['tp'] = insight.TP[-1]
        if insight.SL:
            # sl_points = round(np.abs(insight.SL - request['price']) * si.point)
            # request['sl'] = sl_points if insight.side == IOrderSide.BUY else sl_points
            request['sl'] = insight.SL
        # MT5 requires you to set the position ID if you are wanting to close the order.
        # we can check if it is a close order and set the position ID (open ID is set)
        if insight.order_id:
            # This is likely a close order / reduce order
            request['position'] = insight.order_id
            request["comment"] = f"OlympusTrader Close - {insight.strategyType.value}"

        try:
            check_result = mt5.order_check(request)
            if check_result.retcode  != 0  and check_result.retcode != mt5.TRADE_RETCODE_DONE:
                # TODO: Check if we need to handle this differently return the error
                print(f"Order send failed, retcode={
                      check_result.retcode}, comment={check_result.comment}")
                # retcode=10027 Algorithmic trading is disabled
                #
                return None

            result = mt5.order_send(request)
            if result.retcode != mt5.TRADE_RETCODE_DONE:
                # Should not error here as we have already checked
                print(f"Order send failed, retcode={
                      result.retcode}, comment={result.comment}")
                # retcode=10027 Algorithmic trading is disabled
                #
                return None
            legs = IOrderLegs()
            if result[10].tp:
                legs['take_profit'] = IOrderLeg(
                    order_id=None,
                    limit_price=result[10].tp,
                    filled_price=None,
                    type=IOrderType.LIMIT,
                    status=ITradeUpdateEvent.PENDING_NEW,
                    order_class=IOrderClass.OTO,
                    # created_at=leg.created_at,
                    # updated_at=leg.updated_at,
                    # submitted_at=leg.submitted_at,
                    # filled_at=leg.filled_at
                )
            elif result[10].sl:
                legs['stop_loss'] = IOrderLeg(
                    order_id=None,
                    limit_price=result[10].sl,
                    filled_price=None,
                    type=IOrderType.STOP,
                    status=ITradeUpdateEvent.PENDING_NEW,
                    order_class=IOrderClass.OTO,
                    # created_at=leg.created_at,
                    # updated_at=leg.updated_at,
                    # submitted_at=leg.submitted_at,
                    # filled_at=leg.filled_at
            )
            order = IOrder(
                order_id=result.order,
                asset=asset,
                filled_price=None,
                limit_price=result.ask if insight.side == IOrderSide.BUY else result.bid,
                stop_price=None,
                qty=result.volume,
                filled_qty=None,
                side=insight.side,
                type=insight.type,
                order_class=insight.classType,
                legs=legs
            )
            return order
        except Exception as e:
            print(f"Error: {e}")
            return None

    def startTradeStream(self, callback: Awaitable):
        """
        Starts a listener that will pass the trade events to the strategy
        """
        super().startTradeStream(callback)
        self.RUNNING_TRADE_STREAM = True
        # Rate limit for new trade signals
        rate = 0.5
        loop = asyncio.new_event_loop()

        lastChecked = datetime.now(self.TIMEZONE).replace(tzinfo=timezone.utc)
        while self.RUNNING_TRADE_STREAM:
            sleep(1/rate)
            now = datetime.now(self.TIMEZONE).replace(tzinfo=timezone.utc)
            new_incoming_orders = mt5.history_orders_get(lastChecked, now)
            new_incoming_deals = mt5.history_deals_get(lastChecked, now)
            # TODO:Check if they are sorted!
            if (new_incoming_orders and len(new_incoming_orders) > 0) or (new_incoming_deals and len(new_incoming_deals) > 0):
                # print("Oders", new_incoming_orders)
                # print("Deal", new_incoming_deals)
                for order in new_incoming_orders:
                    try:
                        loop.run_until_complete(callback(ITradeUpdate(order, order.state)))
                    except Exception as e:
                        print(f"Error: {e}")
                        
            lastChecked = now
        loop.close()
# TradeOrder(ticket=530218319, time_setup=1582282114, time_setup_msc=1582282114681, time_done=1582303777, time_done_msc=1582303777582, time_expiration=0, ...

    async def closeTradeStream(self):
        # TODO: Will have to build this out
        self.RUNNING_TRADE_STREAM = False
        pass

    def streamMarketData(self, callback: Awaitable, assetStreams):
        super().streamMarketData(callback, assetStreams)
        self.RUNNING_MARKET_STREAM = True
        barStreamCount = len(
            [asset for asset in assetStreams if asset['type'] == 'bar'])
        # Threading pools for the streams
        pool = ThreadPoolExecutor(max_workers=(
            barStreamCount), thread_name_prefix="MarketDataStream")
        loop = asyncio.new_event_loop()
        for asset in assetStreams:
            try:
                if asset['type'] == 'bar':

                    async def MT5BarStreamer(asset):
                        # Set to last valid time (May need to change as we dont want any signals for past data)
           
                        # lastChecked = pd.Timestamp(mt5.symbol_info(
                        #     asset["symbol"]).time, unit='s', tz=self.TIMEZONE)
                        lastChecked = pd.Timestamp.now(tz=self.TIMEZONE).replace(tzinfo=timezone.utc)
                        while self.RUNNING_MARKET_STREAM:
                            nextTimetoCheck = asset['time_frame'].get_next_time_increment(
                                lastChecked)
                            # .replace(tzinfo=tz)
                            await asyncio.sleep((nextTimetoCheck - lastChecked).total_seconds())
                            # may use self.TF_MAPPING[]

                            # bars = mt5.copy_rates_range(asset['symbol'], int(
                            #     asset['time_frame']), lastChecked.timestamp(), nextTimetoCheck.timestamp())

                            bars = mt5.copy_rates_from(asset['symbol'], int(
                                asset['time_frame']), lastChecked, 1)
                            # bars = mt5.copy_rates_from(asset['symbol'], int(
                            #     asset['time_frame']), lastChecked, 2)
                            # May need the one below if we need to make the bar if the time frame is not supported
                            # bars = mt5.copy_rates_from(asset['symbol'], int(
                            #     asset['time_frame']), lastChecked, asset['time_frame'].amount)
                            if isinstance(bars, np.ndarray) and len(bars) > 0:
                                barDatas = self.format_on_bar(
                                    bars, asset['symbol'])
                                for idx in range(0, len(barDatas)):
                                    try:
                                        bar = barDatas.iloc[[idx]]
                                        # if asset['time_frame'].is_time_increment(bar.index[0][1]) and (not (bar.index[0][1]) < lastChecked):
                                        if asset['time_frame'].is_time_increment(bar.index[0][1]) and (not (bar.index[0][1]) < lastChecked):
                                            # TODO: Handle Feature streams? Strategy already handles renaming the index to include feature name
                                            loop.run_until_complete(
                                                callback(bar, timeframe=asset['time_frame']))
                                    except Exception as e:
                                        print(f"Error: {e}")
                                        continue
                            # Update the last checked time for this stream. -> moved to the top as run_until_complete may take over a min and we are clamped to the next min
                            lastChecked = nextTimetoCheck

                    streamKey = f"{asset['symbol']}:{str(asset['time_frame'])}"
                    self._MARKET_STREAMS[streamKey] = loop.create_task(
                        MT5BarStreamer(asset), name=f"Market:{streamKey}")
                    # self._MARKET_STREAMS[streamKey] = loop.run_in_executor(pool, MT5BarStreamer, asset)
                else:
                    raise NotImplementedError(
                        f"Stream type {asset['type']} not supported")
            except Exception as e:
                print(f"Error: {e}")
                continue
        # TODO: Will need a way to close this stream as its not being referenced anywhere else
        loop.run_forever()

    async def closeStream(self, assetStreams):
        if self.RUNNING_MARKET_STREAM:
            for asset in assetStreams:
                streamKey = f"{asset['symbol']}:{str(asset['time_frame'])}"
                marketStream = self._MARKET_STREAMS.get(streamKey)
                if marketStream:
                    await marketStream.cancel()
                    del self._MARKET_STREAMS[streamKey]
                if len(self._MARKET_STREAMS) == 0:
                    self.RUNNING_MARKET_STREAM = False

    def format_on_bar(self, bar: Any, symbol: Optional[str] = None) -> Optional[pd.DataFrame]:
        if isinstance(bar, np.ndarray):
            index = pd.MultiIndex.from_product(
                [[symbol], pd.to_datetime(bar['time'], utc=True, unit='s')], names=['symbol', 'date'])

            res = pd.DataFrame(data={
                'open': bar['open'],
                'high': bar['high'],
                'low': bar['low'],
                'close': bar['close'],
                'volume': bar['tick_volume'],
            }, index=index, columns=['open', 'high', 'low', 'close', 'volume'])
            return res
        else:
            return None

    def format_on_quote(self, quote: Any, symbol=None) -> IQuote:
        data = IQuote(
            symbol=symbol,
            bid=quote.bid,
            ask=quote.ask,
            bid_size=0,
            ask_size=0,
            volume=quote.volume,
            timestamp=self.time
        )
        return data

    def format_order(self, order: Any) -> IOrder:
        # if isinstance(order, dict):
        #     return order
        if type(order) == IOrder:
            return order
        
        side = None
        orderType = None
        match order.type:
            case mt5.POSITION_TYPE_BUY:
                side = IOrderSide.BUY
                orderType = IOrderType.MARKET
            case mt5.POSITION_TYPE_SELL:
                side = IOrderSide.SELL
                orderType = IOrderType.MARKET
            case mt5.ORDER_TYPE_BUY_LIMIT:
                side = IOrderSide.BUY
                orderType = IOrderType.LIMIT
            case mt5.ORDER_TYPE_SELL_LIMIT:
                side = IOrderSide.SELL
                orderType = IOrderType.LIMIT
            case mt5.ORDER_TYPE_BUY_STOP:
                side = IOrderSide.BUY
                orderType = IOrderType.STOP
            case mt5.ORDER_TYPE_SELL_STOP:
                side = IOrderSide.SELL
                orderType = IOrderType.STOP
            case mt5.ORDER_TYPE_BUY_STOP_LIMIT:
                side = IOrderSide.BUY
                orderType = IOrderType.STOP_LIMIT
            case mt5.ORDER_TYPE_SELL_STOP_LIMIT:
                side = IOrderSide.SELL
                orderType = IOrderType.STOP_LIMIT
            case mt5.ORDER_TYPE_CLOSE_BY:
                pass
            #     side = IOrderSide.BUY
            #     orderType = IOrderType.STOP_LIMIT
            case _:
                raise NotImplementedError("Oder Type not found")

        
            


        orderID = order.ticket
        #  TODO: add support for order legs
        legs = IOrderLegs()
        if order.tp:
            legs['take_profit'] = IOrderLeg(
                order_id=None,
                limit_price=order.tp,
                filled_price=None,
                type=IOrderType.LIMIT,
                status=ITradeUpdateEvent.PENDING_NEW,
                order_class=IOrderClass.OTO,
                created_at=pd.Timestamp(order.time_done, unit='s', tz=timezone.utc),
                # updated_at=leg.updated_at,
                # submitted_at=leg.submitted_at,
                # filled_at=leg.filled_at
            )
        if order.sl:
            legs['stop_loss'] = IOrderLeg(
                order_id=None,
                limit_price=order.sl,
                filled_price=None,
                type=IOrderType.STOP,
                status=ITradeUpdateEvent.PENDING_NEW,
                order_class=IOrderClass.OTO,
                created_at=pd.Timestamp(order.time_done, unit='s', tz=timezone.utc),
                # updated_at=leg.updated_at,
                # submitted_at=leg.submitted_at,
                # filled_at=leg.filled_at
            )
            """
            ticket         
            time_setup  
            time_setup_msc  
            time_expiration  
            type  type_time  type_filling  
            state  magic  volume_current  
            price_open   sl   tp  price_current  symbol 
            comment external_id
            """

        res = IOrder(
            order_id=orderID,
            asset=self.get_ticker_info(order.symbol),
            filled_price=order.price_current if order.price_open != 0.0 else (None if orderType != IOrderType.MARKET else order.price_current),
            limit_price=order.price_open if order.price_open != 0.0 else None,
            stop_price=order.price_stoplimit if order.price_stoplimit != 0.0 else None,
            qty=order.volume_initial + order.volume_current,
            filled_qty=order.volume_initial,
            side=side,
            type=orderType,
            order_class=IOrderClass.SIMPLE if (
                order.tp == 0 and order.sl == 0) else IOrderClass.OTO,
            time_in_force=ITimeInForce.GTC if (
                order.type_time == mt5.ORDER_TIME_GTC) else ITimeInForce.DAY,
            status=order.state,
            created_at=pd.Timestamp(order.time_setup, unit='s', tz=timezone.utc),
            updated_at=pd.Timestamp(order.time_done, unit='s', tz=timezone.utc),
            submitted_at=pd.Timestamp(order.time_setup, unit='s', tz=timezone.utc),
            filled_at= pd.Timestamp(order.time_done, unit='s', tz=timezone.utc),
            legs=legs

        )
        return res

    def format_on_trade_update(self, trade: Any) -> tuple[IOrder, ITradeUpdateEvent]:
        event: ITradeUpdateEvent = None
        match trade.event:
            case mt5.ORDER_STATE_FILLED:
                event = ITradeUpdateEvent.FILLED
            case mt5.ORDER_STATE_PARTIAL:
                event = ITradeUpdateEvent.PARTIAL_FILLED
            case mt5.ORDER_STATE_CANCELED:
                event = ITradeUpdateEvent.CANCELED
            case mt5.ORDER_STATE_REJECTED:
                event = ITradeUpdateEvent.REJECTED
            case mt5.ORDER_STATE_PLACED:
                event = ITradeUpdateEvent.PENDING_NEW
            case mt5.ORDER_STATE_STARTED:
                event = ITradeUpdateEvent.NEW
            case mt5.ORDER_STATE_EXPIRED:
                event = ITradeUpdateEvent.EXPIRED
            case mt5.ORDER_STATE_REQUEST_MODIFY:
                event = ITradeUpdateEvent.REPLACED
            # case mt5.ORDER_STATE_PLACED:
            #     # Might not even be needed  as mt5 does not have the accepted
            #     event = ITradeUpdateEvent.ACCEPTED
            case _:
                event = trade.event
        order = self.format_order(trade.order)
        order["status"] = event
        return order, event



from ..base_executor import BaseExecutor
from ...insight import InsightState
from ....broker.interfaces import IOrderSide
from ....strategy.interfaces import IStrategyMode


class BasicStopLossExecutor(BaseExecutor):
    """ 
    ### Executor for Basic Stop Loss

    This executor is used to check if the insight has crossed the stop loss level. If the price crosses the stop loss level, the position is closed.
    Ideally this executor should be used in conjunction simple orders types with no OCO or bracket orders as the broker will close the position if the stop loss level is crossed.

    :param strategy (BaseStrategy): The strategy instance

    Note: You should have already set the stop loss price in the insight before using this executor.

    Author:
        @isaac-diaby
    """

    def __init__(self, strategy):
        super().__init__(strategy, InsightState.FILLED, "1.0")

    def run(self, insight):
        #  Check if the insight already has a stop loss order leg
        if insight.stopLossOrderLeg:
            return self.returnResults(True, True, "Insight already has a stop loss order")
        # Check if the insight has a stop loss price
        if insight.SL == None:
            return self.returnResults(True, True, "Insight does not have a stop loss price set.")
        try:
            # check if the insight has not already been closed 
            if insight._closing:
                return self.returnResults(False, True, "Insight is being closed.")
            # Check if price broke the stop loss level
            latestBar = self.get_latest_bar(insight.symbol)
            latestQuote = self.get_latest_quote(insight)
            shouldClose = False
            match insight.side:
                case IOrderSide.BUY:
                    if (latestBar.low < insight.SL) or (latestQuote['bid'] < insight.SL):
                        shouldClose = True
                case IOrderSide.SELL:
                    if (latestBar.high > insight.SL) or (latestQuote['ask'] > insight.SL):
                        shouldClose = True
            if shouldClose:
                if self.STRATEGY.insights[insight.INSIGHT_ID].close():
                    return self.returnResults(False, True, f"Price broke the stop loss level: {insight.symbol} : {insight.SL}. Closing position.")
                return self.returnResults(False, False, f"Error closing position.")

            return self.returnResults(True, True, f"Stop loss price has not been reached: {insight.symbol} : {insight.SL}")
        except Exception as e:
            return self.returnResults(False, False, f"Error Checking Stop Loss: {e}")



from ..base_executor import BaseExecutor
from ...insight import InsightState


class MinimumRiskToRewardExecutor(BaseExecutor):
    """
    ### Executor for Minimum Risk to Reward Ratio
    This executor is used reject insights that do not meet the minimum risk to reward ratio (RRR).

    **This will update your strategy instance with the minimum risk to reward ratio required.** (STRATEGY.minRewardRiskRatio), if not already set.

    :param strategy (BaseStrategy): The strategy instance
    :param minimumRR (float): The minimum risk to reward ratio required. Default is 2.0

    Note: You will need to have already set the limit price, stop loss and  take profit levels in the insight before using this executor.
    
    Author:
        @isaac-diaby
    """

    def __init__(self, strategy, minimumRR=None):
        super().__init__(strategy, InsightState.NEW, "1.0")

        # Set the minimum risk to reward ratio required for the strategy if provided
        if minimumRR is not None:
            self.STRATEGY.minRewardRiskRatio = minimumRR

        # check that the minimum risk to reward ratio is set
        if self.STRATEGY.minRewardRiskRatio is None or self.STRATEGY.minRewardRiskRatio <= 0:
            raise ValueError(
                "Please set the minimum risk to reward ratio required for the strategy")

    def run(self, insight):
        # Check if the insight has a limit price, stop loss and take profit levels
        if insight.limit_price is None or insight.SL is None or insight.TP is None:
            return self.returnResults(False, False, "Insight does not have limit price, stop loss or take profit levels set.")
        minimumRR = self.STRATEGY.minRewardRiskRatio
        RR = insight.getPnLRatio()
        if RR < minimumRR:
            response = self.returnResults(
                False, True, f"Risk to Reward ratio (RRR) is less than the minimum required -  {RR} < {minimumRR}")
            self.changeState(insight, InsightState.REJECTED, response.message)
            return response

        return self.returnResults(True, True)



class AttributeDict(dict):
    """
    This is adds functions to the dictionary class, no other modifications. This gives dictionaries abilities like:
    print(account.BTC) -> {'available': 1, 'hold': 0}
    account.BTC = "cool"
    print(account.BTC) -> cool
    Basically you can get and set attributes with a dot instead of [] - like dict.available rather than
     dict['available']
    """

    def __getattr__(self, attr):
        # Try catch is wrapped to support copying objects
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(attr)

    def __setattr__(self, attr, value):
        self[attr] = value



import dash
from dash import dcc, html
from dash.dependencies import Input, Output
import datetime

# insight = {
#     "INSIGHT_ID": "12345",
#     "PARENT": None,
#     "order_id": "54321",
#     "side": "buy",
#     "symbol": "AAPL",
#     "quantity": 100,
#     "contracts": 10,
#     "type": "market",
#     "classType": "equity",
#     "limit_price": 150.0,
#     "TP": [155.0, 160.0],
#     "SL": 145.0,
#     "strategyType": "momentum",
#     "confidence": 0.85,
#     "tf": "1D",
#     "state": "new",
#     "createAt": "2023-10-01T12:00:00Z",
#     "updatedAt": "2023-10-02T12:00:00Z",
#     "filledAt": None,
#     "closedAt": None,
#     "close_price": None
# }

def InsightDetails(insight):
    def format_date(date_string):
        if not date_string:
            return 'N/A'
        return datetime.datetime.fromisoformat(date_string.replace("Z", "+00:00")).strftime('%Y-%m-%d %H:%M:%S')
    
    def get_state_icon(state):
        icons = {
            'new': '🕒',
            'filled': '✅',
            'cancelled': '⚠️',
            'closed': '✅'
        }
        return icons.get(state.lower(), '')
    
    def detail_item(label, value, icon=None):
        return html.Div([
            html.Span(f"{label}:", className="text-[#be9e6b]"),
            html.Span([
                html.Span(icon, className="mr-2") if icon else '',
                html.Span(value, className="text-white")
            ], className="flex items-center")
        ], className="flex items-center justify-between border-b border-[#be9e6b] py-2")

    return  html.Div([
            html.Div([
                html.H2("Insight Details", className="text-2xl font-bold text-[#be9e6b] mb-4"),
                html.Div([
                    detail_item("Insight ID", insight["INSIGHT_ID"]),
                    detail_item("Parent ID", insight["PARENT"] or 'N/A'),
                    detail_item("Order ID", insight["order_id"] or 'N/A'),
                    detail_item("Side", insight["side"] or 'N/A', icon='📈' if insight["side"] == 'buy' else '📉'),
                    detail_item("Symbol", insight["symbol"] or 'N/A'),
                    detail_item("Quantity", str(insight["quantity"]) or 'N/A'),
                    detail_item("Contracts", str(insight["contracts"]) or 'N/A'),
                    detail_item("Type", insight["type"] or 'N/A'),
                    detail_item("Class Type", insight["classType"] or 'N/A'),
                    detail_item("Limit Price", str(insight["limit_price"]) or 'N/A'),
                    detail_item("Take Profit", ', '.join(map(str, insight["TP"])) if insight["TP"] else 'N/A'),
                    detail_item("Stop Loss", str(insight["SL"]) or 'N/A'),
                    detail_item("Strategy Type", insight["strategyType"] or 'N/A'),
                    detail_item("Confidence", f"{insight['confidence'] * 100:.2f}%" if insight["confidence"] else 'N/A'),
                    detail_item("Timeframe", insight["tf"] or 'N/A'),
                    detail_item("State", insight["state"], icon=get_state_icon(insight["state"])),
                    detail_item("Created At", format_date(insight["createAt"]), icon='🕒'),
                    detail_item("Updated At", format_date(insight["updatedAt"]), icon='🕒'),
                    detail_item("Filled At", format_date(insight["filledAt"]), icon='🕒'),
                    detail_item("Closed At", format_date(insight["closedAt"]), icon='🕒'),
                    detail_item("Close Price", str(insight["close_price"]) or 'N/A')
                ], className="grid grid-cols-1 md:grid-cols-2 gap-4")
            ], className="bg-[#0a0f1a] rounded-lg shadow-lg border border-[#be9e6b] p-6")
        ])
       



from ..base_executor import BaseExecutor
from ...insight import InsightState
import numpy as np


class FullAccountQuantityToRiskExecutor(BaseExecutor):
    """
    ### Executor for Full Account Quantity to Risk
    This executor uses the full account cash balance to determine the quantity to trade.
    
    :param strategy (BaseStrategy): The strategy instance

    #### How it works:
    When a new insight is generated, the executor calculates the quantity to trade based on the account balance.

    Note: limit price ust be set in the insight before using this executor.

    Author:
        @isaac-diaby
    """

    def __init__(self, strategy):
        super().__init__(strategy, InsightState.NEW, "1.0")

    def run(self, insight):
        if insight.limit_price is None or insight.SL is None:
            return self.returnResults(False, False, "Insight does not have limit price or stop loss levels set.")

        # Calculate the quantity to trade
        try:
            # Account size at to place order
            account_size_at_risk = self.STRATEGY.account.buying_power if self.STRATEGY.assets[insight.symbol]['marginable'] else self.STRATEGY.account.cash
            if self.STRATEGY.broker.supportedFeatures.maxOrderValue is not None:
                # Check if the account size at risk is greater than the maximum order value supported by the broker
                account_size_at_risk = min(self.STRATEGY.broker.supportedFeatures.maxOrderValue, account_size_at_risk)

            # Calculate the quantity to trade based on the limit price and account size cash balance
            size_should_buy = account_size_at_risk/insight.limit_price


            # Round down the quantity to the nearest whole number
            if size_should_buy > 1:
                size_should_buy = np.floor(size_should_buy)

            else:
                # Check if the quantity is greater than the minimum order size
                if size_should_buy < self.STRATEGY.assets[insight.symbol]['min_order_size']:
                    response = self.returnResults(False, False, f"Quantity is less than the minimum order size: {
                                                self.STRATEGY.assets[insight.symbol]['min_order_size']}")
                    self.changeState(
                        insight, InsightState.REJECTED, response.message)
                    return response
                pass

            # Update the quantity in the insight
            self.STRATEGY.insights[insight.INSIGHT_ID].update_quantity(
                size_should_buy)

            return self.returnResults(True, True, f"Quantity set to {size_should_buy}")
        except Exception as e:
            return self.returnResults(False, False, f"Error calculating quantity: {e}")



from concurrent.futures import ThreadPoolExecutor
import os
from typing import Any, Optional, Union

from ccxt.base.exchange import Exchange
from ccxt.base.types import Order, OrderSide, Ticker
import pandas as pd

from OlympusTrader.broker.interfaces import IQuote
from OlympusTrader.insight.insight import Insight

from .base_broker import BaseBroker
from .interfaces import IAccount, IAsset, IOrder, IOrderLeg, IOrderLegs, IOrderSide, IOrderType, IPosition, ISupportedBrokerFeatures, ISupportedBrokers, ITradeUpdate, ITradeUpdateEvent
import asyncio


class CCXTBroker(BaseBroker):
    """CCXT
    CCXT Broker.
    """

    exchange: Exchange
    """Yout CCXT Exchange. instance of ccxt.Exchange"""

    def __init__(self, exchange: Exchange,  paper: bool, feed=None):

        # FIXME: The broker is not yet supported by the broker manager
        # - considering not even using CCXT and just exchange rest api and websocket directly as we have custome implementations for each broker
        # - this will allow us to have more control over the broker and not rely on ccxt
        
        raise NotImplementedError("CCXT Broker is not yet supported by the broker manager")
    
        assert issubclass(
            type(exchange), Exchange), 'exchange must be a subclass of ccxt.Exchange'
        self.exchange = exchange
        if paper:
            if exchange.has['sandbox']:
                self.exchange.set_sandbox_mode(True)
            else:
                raise Exception(
                    'This exchange does not support sandbox mode', 'sandbox')
        super().__init__(f"{ISupportedBrokers.CCXT}_{
            self.exchange.id}", paper, feed)
        

    def get_ticker_info(self, symbol: str) -> Union[IAsset, None]:
        try:
            if symbol in self.TICKER_INFO:
                return self.TICKER_INFO[symbol]

            if self.exchange.has['fetchMarkets']:
                loop = asyncio.new_event_loop()
                markets = loop.run_until_complete(self.exchange.load_markets())
                # symbol = symbol.replace('-', '/')
                tickerInfo = markets.get(symbol)

                if not tickerInfo:
                    raise Exception(
                        'This exchange does not support the symbol', symbol)
            

                # CCXT doesnt really have stocks so we will just assume its a crypto
                tickerAsset: IAsset = IAsset(
                    id=tickerInfo['id'],
                    symbol=tickerInfo['symbol'],
                    name=tickerInfo['info']['fullName'],
                    asset_type='crypto',
                    exchange=self.exchange.id,
                    status="active" if tickerInfo['active'] else "inactive",
                    tradable=tickerInfo['active'],
                    marginable=tickerInfo['margin'],
                    shortable=True,
                    # shortable=tickerInfo['info']['shortable'],
                    fractionable=True,
                    # fractionable=tickerInfo['info']['fractionable'],
                    min_order_size=tickerInfo['limits']['amount']['min'],
                    min_price_increment=tickerInfo['precision']['price']
                )
                self.TICKER_INFO[symbol] = tickerAsset
                return tickerAsset
            else:
                raise Exception('This exchange does not support load_markets')
        except Exception as e:
                print(f"Error: {e}")
                return None

    def get_account(self):
        try:
            if self.exchange.has['fetch_balance']:
                res = self.exchange.fetch_balance()
                account: IAccount = IAccount(account_id=str(res.id), equity=float(res.equity), cash=float(res.cash), currency=res.currency,
                                             buying_power=float(res.buying_power), leverage=float(res.multiplier), shorting_enabled=res.shorting_enabled)
            else:
                # TODO: Implement this for alpaca etc
                raise Exception('This exchange does not support fetch_balance')
        except Exception as e:
            print(f"Error: {e}")
            return None

    def get_position(self, symbol: str):
        try:
            if self.exchange.has['fetch_position']:
                res = self.exchange.fetch_position(symbol)
                position: IPosition = IPosition(asset=self.get_ticker_info(symbol), avg_entry_price=float(res.avg_entry_price), qty=float(res.qty), side=res.side, market_value=float(
                    res.market_value), cost_basis=float(res.cost_basis), current_price=float(res.current_price), unrealized_pl=float(res.unrealized_pl))
            else:
                raise Exception(
                    'This exchange does not support fetch_position')
        except Exception as e:
            print(f"Error: {e}")
            return None

    def format_position(self, position: Any) -> IPosition:
        try:
            return IPosition(
                asset=self.get_ticker_info(position.symbol),
                avg_entry_price=float(position.avg_entry_price),
                qty=float(position.qty),
                side=position.side.value,
                market_value=float(position.market_value),
                cost_basis=float(position.cost_basis),
                current_price=float(position.current_price),
                unrealized_pl=float(position.unrealized_pl)
            )
        except Exception as e:
            print(f"Error: {e}")
            return None

    def get_history(self, asset, start, end, resolution):
        try:
            if self.exchange.has['fetch_ohlcv']:
                history = self.exchange.fetch_ohlcv(
                    asset.symbol, resolution.value, start, end)
                return self.format_on_bar(history, asset.symbol)
            else:
                raise Exception('This exchange does not support fetch_ohlcv')
        except Exception as e:
            print(f"Error: {e}")
            return None

    def format_on_bar(self, bar: Any, symbol: str) -> dict:
        # if bar is None:
        #     return None

        if isinstance(bar, list[list]):
            # [timestamp, open, high, low, close, volume]?
            pass
            # index = pd.MultiIndex.from_product(
            #     [[bar.symbol], [bar.timestamp]], names=['symbol', 'date'])
            # data = pd.DataFrame(data={
            #     'open': bar.open,
            #     'high': bar.high,
            #     'low': bar.low,
            #     'close': bar.close,
            #     'volume': bar.volume,
            # }, index=index, columns=['open', 'high', 'low', 'close', 'volume'])
        # data.set_index(['symbol', )], inplace=True)
            # return data
        elif isinstance(bar, list[list]):
            # [[timestamp], [open], [high], [low], [close], [volume]]
            pass

        elif bar is None:
            return None
        else:
            raise Exception('Cant format bar')

    def get_position(self, symbol: str):
        try:
            if self.exchange.has['fetch_position']:
                res = self.exchange.fetch_position(symbol)
                position: IPosition = IPosition(asset=self.get_ticker_info(symbol), avg_entry_price=float(res.avg_entry_price), qty=float(res.qty), side=res.side, market_value=float(
                    res.market_value), cost_basis=float(res.cost_basis), current_price=float(res.current_price), unrealized_pl=float(res.unrealized_pl))
            else:
                raise Exception(
                    'This exchange does not support fetch_position')
        except Exception as e:
            print(f"Error: {e}")
            return None

    def get_positions(self):
        try:
            if self.exchange.has['fetch_positions']:
                res = self.exchange.fetch_positions()
                positions: dict[str, IPosition] = []
                for position in res:
                    positions.append(self.format_position(position))
                return positions
            else:
                raise Exception(
                    'This exchange does not support fetch_positions')
        except Exception as e:
            print(f"Error: {e}")
            return None

    def format_position(self, position: Any) -> IPosition:
        print("Position:", position)
        pos = IPosition(
            asset=self.get_ticker_info(position.symbol),
            avg_entry_price=float(position.avg_entry_price),
            qty=float(position.qty),
            side=position.side.value,
            market_value=float(position.market_value),
            cost_basis=float(position.cost_basis),
            current_price=float(position.current_price),
            unrealized_pl=float(position.unrealized_pl)
        )
        return pos

    def get_orders(self):
        try:
            if self.exchange.has['fetch_orders']:
                res = self.exchange.fetch_orders()
                orders: dict[str, IOrder] = []
                for order in res:
                    orders.append(self.format_order(order))
                return orders
            else:
                raise Exception(
                    'This exchange does not support fetch_orders')
        except Exception as e:
            print(f"Error: {e}")
            return None

    def get_order(self, order_id: str):
        try:
            if self.exchange.has['fetch_order']:
                res = self.exchange.fetch_order(order_id)
                return self.format_order(res)
            else:
                raise Exception(
                    'This exchange does not support fetch_order')
        except Exception as e:
            print(f"Error: {e}")
            return None

    def format_order(self, order: Order) -> IOrder:
        try:
            if order is None:
                return None
            if isinstance(order, Order):
                legs = IOrderLegs()
                if order['takeProfitPrice']:
                    print(order['trades'])
                    legs["take_profit"] = IOrderLeg(
                        limit_price=order['takeProfitPrice'],
                        filled_price=None,
                        type=IOrderType.LIMIT,
                        status=None,
                        order_class=None,
                        created_at=None,
                        updated_at=None,
                        submitted_at=None,
                        filled_at=None
                    )
                if order['stopLossPrice']:
                    print(order['trades'])
                    legs["stop_loss"] = IOrderLeg(
                        limit_price=order['stopLossPrice'],
                        filled_price=None,
                        type=IOrderType.STOP,
                        status=None,
                        order_class=None,
                        created_at=None,
                        updated_at=None,
                        submitted_at=None,
                        filled_at=None
                    )
                return IOrder(
                    order_id=order["clientOrderId"],
                    asset=self.get_ticker_info(order["symbol"]),
                    side=IOrderSide.BUY if (
                        order["side"] == 'buy') else IOrderSide.SELL,
                    qty=float(order["amount"]),
                    type=order["type"],
                    order_class=order["trades"][0]["type"],
                    time_in_force=order["timeInForce"],
                    status=order["status"],
                    limit_price=float(order["price"]),
                    stop_price=float(order["stopPrice"]),
                    filled_price=float(order["average"]),
                    filled_qty=float(order["filled"]),
                    created_at=order["datetime"],
                    updated_at=order["lastUpdateTimestamp"],
                    filled_at=order["lastTradeTimestamp"],
                    submitted_at=order["timestamp"],
                    extra=order["info"],
                )
        except Exception as e:
            print(f"Error: {e}")
            return None

    def get_latest_quote(self, asset: IAsset) -> IQuote:
        try:
            if self.exchange.has['fetch_ticker']:
                res = self.exchange.fetch_ticker(asset.symbol)
                return IQuote(
                    symbol=asset.symbol,
                    ask=float(res['ask']),
                    bid=float(res['bid']),
                    last=float(res['last'])
                )
            else:
                raise Exception('This exchange does not support fetch_ticker')
        except Exception as e:
            print(f"Error: {e}")
            return None

    def format_on_quote(self, quote: Ticker) -> IQuote:
        try:
            return IQuote(
                symbol=quote.symbol,
                ask=float(quote['ask']),
                ask_size=float(quote['askVolume']),
                bid=float(quote['bid']),
                bid_size=float(quote['bidVolume']),
                volume=float(quote['quoteVolume']),
                timestamp=quote['timestamp']
            )
        except Exception as e:
            print(f"Error: {e}")
            return None

    def execute_insight_order(self, insight: Insight, asset: IAsset) -> IOrder | None:
        super().execute_insight_order(insight, asset)
        order_params = {
            'symbol': asset.symbol,
            'type': insight.type.value,
            'side':  'buy' if insight.side == IOrderSide.BUY else 'sell',
            'amount': insight.quantity,
            'params': {}
        }
        try:
            order_params['type'] = 'market' if insight.type == IOrderType.MARKET else 'limit'
            if insight.type == IOrderType.LIMIT:
                order_params['price'] = insight.limit_price

            if insight.type == IOrderType.STOP or insight.type == IOrderType.STOP_LIMIT:
                order_params['params']['stopPrice'] = insight.limit_price
            else:
                if insight.TP and insight.SL and self.exchange.has['createOrderWithTakeProfitAndStopLoss']:
                    order_params['params'] = self.exchange.set_take_profit_and_stop_loss_params(
                        symbol=asset.symbol, side=order_params['side'], amount=order_params['amount'], price=order_params['price'], takeProfit=insight.TP[-1], stopLoss=insight.SL, params=order_params['params'])

                elif insight.TP:
                    assert self.exchange.has['createTakeProfitOrder'], 'This exchange does not support createTakeProfitOrder'
                    order_params['params']['takeProfitPrice'] = insight.TP[-1]
                elif insight.SL:
                    assert self.exchange.has['createStopLossOrder'], 'This exchange does not support createStopLossOrder'
                    order_params['params']['stopLossPrice'] = insight.SL
        except Exception as e:
            print(f"Error composing order: {e}")
            return None
        try:
            order = self.exchange.create_order(**order_params)
            return self.format_order(order)
        except Exception as e:
            print(f"Error sending order: {e}")
            raise e

    def cancel_order(self, order_id: str):
        try:
            if self.exchange.has['cancelOrder']:
                cancelled_order = self.exchange.cancel_order(order_id)
                return self.format_order(cancelled_order)
            else:
                raise Exception('This exchange does not support cancelOrder')
        except Exception as e:
            print(f"Error cancelling order: {e}")
            return None

    def close_all_positions(self):
        # close all positions
        self.exchange.close_all_positions()
        pass

    def close_position(self, symbol: str, qty=None, percent=None):
        # Implement logic to close a specific position
        raise NotImplementedError()
        pass

    def format_on_trade_update(self, trade: Union[ITradeUpdate, Any]):
        if isinstance(trade, ITradeUpdate):
            return super().format_on_trade_update(trade)
        event: ITradeUpdateEvent = None
        match trade['status']:
            case 'open':
                event = ITradeUpdateEvent.NEW
            case 'closed':
                event = ITradeUpdateEvent.FILLED
            case 'canceled':
                event = ITradeUpdateEvent.CANCELED
            case 'rejected':
                event = ITradeUpdateEvent.REJECTED
            case 'expired':
                event = ITradeUpdateEvent.EXPIRED
        return self.format_order(trade.order), event

    def startTradeStream(self, callback):
        # Implement logic to start trade stream
        super().startTradeStream(callback)
        self.RUNNING_TRADE_STREAM = True
        while self.RUNNING_TRADE_STREAM:
            if self.exchange.has['watchMyTradesForSymbols']:
                trades = self.exchange.watch_my_trades_for_symbols(
                    self.TICKER_INFO.keys())
            elif self.exchange.has['watchMyTrades']:
                for symbol in self.TICKER_INFO.keys():
                    trades = self.exchange.watch_my_trades(symbol)
            for trade in trades:
                tradeUpdate = self.format_on_trade_update(trade)
                if tradeUpdate:
                    callback(tradeUpdate)

    def streamMarketData(self, callback, assetStreams):
        super().streamMarketData(callback, assetStreams)
        barStreamCount = len(
            [asset for asset in assetStreams if asset['type'] == 'bar'])

        pool = ThreadPoolExecutor(max_workers=(
            barStreamCount), thread_name_prefix="MarketDataStream")
        loop = asyncio.new_event_loop()

        self.RUNNING_MARKET_STREAM = True
        while self.RUNNING_MARKET_STREAM:

            for assetStream in assetStreams:
                match assetStream['type']:
                    case 'bar':
                        async def streamBarDataHandler(assetStream, callback):
                            while self.RUNNING_MARKET_STREAM:
                                barData = self.format_on_bar(await self.exchange.watch_ohlcv(assetStream['symbol'], assetStream['time_frame']))
                                timestamp = barData.index[0][1]
                                if assetStream['time_frame'].is_time_increment(timestamp):
                                    await callback(barData, timeframe=assetStream['time_frame'])
                        # self.exchange.watch_ohlcv(assetStream['symbol'], assetStream['time_frame'])
                        pool.submit(streamBarDataHandler,
                                    assetStream, callback)

                    # case 'quote':
                    #     pool.submit(self.streamQuoteData, assetStream, callback)
                    # case 'trade':
                    #     pool.submit(self.streamTradeData, assetStream, callback)
                    case _:
                        pass

        pass

    async def closeStream(self, assetStreams):
        # Implement logic to close market data stream
        self.RUNNING_MARKET_STREAM = False

    async def closeTradeStream(self):
        # Implement logic to close trade stream
        self.RUNNING_TRADE_STREAM = False



import dash
from dash import dcc, html, callback, no_update
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import pandas as pd
from dash_tvlwc import Tvlwc
from dash_tvlwc.types import SeriesType, ColorType, LineType

from OlympusTrader.broker.interfaces import IAsset
from OlympusTrader.ui.components.dashboardNav import subnav
from OlympusTrader.ui.components.matricItem import MatricItem
from OlympusTrader.ui.components.tradesTable import tradeTable
from OlympusTrader.ui.interfaces.store import STRATEGY_STORE_MAPPINGS


dash.register_page(__name__, title="OlympusTrader - Strategy Overview", )

# assets = [
#     {"symbol": "BTC/USD", "timeframe": "1h", "data": genMockDataFrame(365, 1.2000, 'BTC/USD', '19/3/2023', seed=41), "winrate": "78%", "profit_factor": "2.1",
#      "sharpe_ratio": "1.8", "max_drawdown": "2.5%", "avg_win_amount": "£250.75", "avg_loss_amount": "£80.59", "avg_win_percent": "2.15%", "avg_loss_percent": "1.05%"},
#     {"symbol": "ETH/USD", "timeframe": "1h", "data": genMockDataFrame(365, 1.2000, 'ETH/USD', '19/3/2023', seed=42), "winrate": "65%", "profit_factor": "1.9",
#      "sharpe_ratio": "1.6", "max_drawdown": "3.0%", "avg_win_amount": "£220.50",  "avg_loss_amount": "£92.20", "avg_win_percent": "2.00%", "avg_loss_percent": "1.10%"},
#     {"symbol": "XRP/USD", "timeframe": "15m", "data": genMockDataFrame(365, 1.2000, 'XRP/USD', '19/3/2023', seed=43), "winrate": "52%", "profit_factor": "2.3",
#      "sharpe_ratio": "2.0", "max_drawdown": "2.0%", "avg_win_amount": "£180.25",  "avg_loss_amount": "£30.48", "avg_win_percent": "1.85%", "avg_loss_percent": "0.95%"},
#     {"symbol": "ADA/USD", "timeframe": "1d", "data": genMockDataFrame(365, 1.2000, 'ADA/USD', '19/3/2023', seed=44), "winrate": "81%", "profit_factor": "1.7",
#      "sharpe_ratio": "1.4", "max_drawdown": "3.5%", "avg_win_amount": "£200.00",  "avg_loss_amount": "£340.11", "avg_win_percent": "1.95%", "avg_loss_percent": "1.20%"}
# ]


def assetCard(asset: IAsset, idx, isSelected=False):
    className = "p-4 rounded-lg cursor-pointer transition-colors "
    className += "bg-accent text-primary " if isSelected else "bg-primary-light text-white hover:bg-accent hover:text-primary"
    return html.Div(
        id={'type': 'asset-card', 'index': idx},
        className=className,
        children=[
            html.Div([
                html.P(asset['symbol'], className="text-sm text-white")
            ], className="flex items-center justify-between"),
            html.P(asset['status'], className="mt-2 text-sm")
        ]
    )


assets_in_play_section = html.Div(
    className="bg-primary-foreground rounded-lg shadow-lg border border-accent p-6 mb-8",
    children=[
        html.H2("Assets in Play",
                className="text-2xl font-semibold text-accent mb-4"),
        html.Div(
            className="grid grid-cols-1 md:grid-cols-2 lg:grid-cols-4 gap-4",
            # children=[
            #     assetCard(asset=asset, idx=idx) for idx, asset in enumerate(assets)
            # ],
            id='assets-in-play'
        ),
        dcc.Store(id='selected-asset-store')
    ]
)


@callback(
    Output('assets-in-play', 'children'),
    [Input(STRATEGY_STORE_MAPPINGS.assets.id, 'data')],
    State('assets-in-play', 'children'),

    # prevent_initial_call=True
)
def update_assets_in_play(assets, children):
    if children != None:
        if not assets and len(children) > 0:
            return []
        elif len(assets) == len(children):
            return no_update

    if not assets:
        return no_update

    return [assetCard(asset, idx) for idx, asset in assets.items()]


def assetChart():
    try:
        # copy = data.copy()
        # copy.reset_index(inplace=True)
        # copy['time'] = copy['time'].dt.strftime('%Y-%m-%d')
        # ohlcvt = copy.to_dict('records')
        return Tvlwc(
            id='asset-chart',
            # seriesData=[ohlcvt],
            seriesTypes=[SeriesType.Candlestick],
            width='100%',
            chartOptions={
                'layout': {
                    'background': {'type': ColorType.Solid, 'color': '#0a0f1a'},
                    'textColor': '#fffffa',
                },
                'grid': {
                    'vertLines': {'visible': True, 'style': 0, 'color': '#be9e6b40'},
                    'horzLines': {'visible': True, 'style': 0, 'color': '#be9e6b40'},
                },
                'timeScale': {'timeVisible': True, 'visible': True},

                'localization': {
                    'timeFormatter': "businessDayOrTimestamp => {return Date(businessDayOrTimestamp);}",
                 
                #     'locale': 'en-US',
                #     'priceFormatter': "(function(price) { return '$' + price.toFixed(2); })"
                }
            },
            fullTimeScaleOptions={
                'visible': True,
                'timeVisible': True,
            }
        )
    except Exception as e:
        print(e)
        return None
    
@callback(
    Output("asset-chart", "seriesData"),
    [ 
        Input('selected-asset-store', 'data'),
        Input(STRATEGY_STORE_MAPPINGS.history.id, 'data')
    ],
    State("asset-chart", "seriesData"),
    prevent_initial_call=True,
    suppress_callback_exceptions=True
)
def update_asset_chart(selected_asset, history, seriesData):
    if selected_asset is None or history is None:
        raise PreventUpdate
    
    if selected_asset not in history:
        raise PreventUpdate
    
    # No history data
    if len(history[selected_asset]) == 0:
        return no_update
    
    if seriesData == None:
        if selected_asset not in history:
            return no_update
        else:
            seriesData = [[]]
            seriesData[0].extend(history[selected_asset])
            return seriesData
            
    if (len(seriesData[0]) == len(history[selected_asset])) or seriesData[0][-1]["time"] == history[selected_asset][-1]["time"]:
        return no_update
    
    indexdiff = len(history[selected_asset]) - len(seriesData[0])
    seriesData[0].extend(history[selected_asset][-indexdiff:])

    print(history[selected_asset][-1])
    print(seriesData[0][-1])

    return seriesData


def strategy_metrics_section(asset: IAsset):
    return html.Div(
        className="bg-primary-foreground rounded-lg shadow-lg border border-accent p-6",
        children=[
            html.H3("Asset Info",
                    className="text-xl font-semibold text-accent mb-4"),
            html.Div(
                className="grid grid-cols-2 gap-4",
                children=[
                    MatricItem(label="Name", value=asset['name']),
                    MatricItem(label="Status",
                               value=asset['status'].capitalize()),
                    MatricItem(label="Tadable",
                               value="True" if asset['tradable'] else "False"),
                    MatricItem(label="Shortable",
                               value="True" if asset['shortable'] else "False"),
                    MatricItem(label="Marginable",
                               value="True" if asset['marginable'] else "False"),
                    MatricItem(
                        label="Fractionable", value="True" if asset['fractionable'] else "False"),
                    MatricItem(label="Min Order Size",
                               value=asset.get('min_order_size', "N/A")),
                    MatricItem(label="Max Order Size",
                               value=asset.get('max_order_size', "N/A"))
                ]
            )

        ]
    )


def chart_section(asset: IAsset):
    return html.Div(
        children=[
            html.H2(f"{asset['symbol']} - Chart",
                    className="text-2xl font-semibold text-accent mb-4"),
            html.Div(
                className="relative min-h-64 bg-primary-light rounded flex items-center justify-center",
                children=assetChart()
                # [
                    # assetChart(assets[idx]['data'])
                    # (html.P(
                    #     className="text-accent",
                    #     children="Chart Placeholder"
                    # ) if asset is None else assetChart(),
                    # ) if asset is None else assetChart(assets(['data'])),
                    # strategy_metrics_section(idx)
                # ]
            )
        ]
    )


# Define the layout
layout = html.Div([
    subnav(),
    html.Div(
        className="flex-grow container mx-auto py-8 text-white",
        children=[
            html.H1('Assets Overview',
                    className="text-3xl font-bold text-accent mb-6"),
            assets_in_play_section,
            html.Div(id='selected-asset-chart',
                     className="bg-primary-foreground rounded-lg shadow-lg border border-accent p-6", children=[]),
            html.Div(id='selected-asset-metrics',
                     className="mt-8 grid grid-cols-1 md:grid-cols-2 gap-6", children=[])
        ]
    )
])

# Callback to update the selected-asset-store and display the chart section


@callback(
    [Output('selected-asset-store', 'data'),
     Output('selected-asset-chart', 'children'),
     Output('selected-asset-metrics', 'children')
     ],
    [
        Input({'type': 'asset-card', 'index': dash.dependencies.ALL}, 'n_clicks'),
    ],
    [State(STRATEGY_STORE_MAPPINGS.assets.id, 'data'),
     State('selected-asset-store', 'data')]
)
def update_selected_asset_store(n_clicks, assets: IAsset, selected_asset_data):
    ctx = dash.callback_context

    if not ctx.triggered or not n_clicks or all(click is None for click in n_clicks):
        return None, no_update, []

    # Get the index of the clicked asset card
    prop_id = ctx.triggered[0]['prop_id']
    clicked_index = str(prop_id.split('.')[0].split(':')[
                        1].split(',')[0].strip('"'))

    if assets.get(clicked_index, None) is None:
        return None, f"No Data Found For Asset: {clicked_index}", []

    display_chart = no_update
    # Update the selected asset data
    if selected_asset_data is None or selected_asset_data != clicked_index:
        selected_asset_data = clicked_index
        display_chart = chart_section(assets[clicked_index])
    else:
        selected_asset_data = no_update

    return selected_asset_data, display_chart, [strategy_metrics_section(assets[clicked_index]), tradeTable("strategy", FILTERS={"symbol": assets[clicked_index]['symbol']})]

# Callback to update the class names of all asset cards


@callback(
    Output({'type': 'asset-card', 'index': dash.dependencies.ALL}, 'className'),
    [Input('selected-asset-store', 'data')],
    [State(STRATEGY_STORE_MAPPINGS.assets.id, 'data')]
)
def update_class_names(selected_asset_data, assets):
    # default class names for the asset cards
    class_names = [
        "p-4 bg-primary-light rounded-lg shadow-md cursor-pointer bg-primary-light text-white hover:bg-accent hover:text-primary"] * len(assets)
    
    if selected_asset_data is None:
        # Return the default class names
        return class_names

    # Update the class name of the selected asset card
    class_names[list(assets).index(selected_asset_data)
                ] = "p-4 bg-primary-light rounded-lg shadow-md cursor-pointer border border-accent"

    return class_names



from typing import override


from .base_strategy import BaseStrategy
from ..broker.base_broker import BaseBroker
from ..insight.insight import Insight



class Strategy(BaseStrategy):
    def __init__(self, broker: BaseBroker, **kwargs):
        super().__init__(broker, **kwargs)

    @override
    def start(self):
        super().start()

    @override
    def init(self, asset):
        super().init(asset)

    @override
    def universe(self) -> set[str]:
        super().universe()
        return {"AAPL", "MSFT", "GOOGL", "AMZN", "TSLA"}

    @override
    def on_bar(self, symbol, bar):
        super().on_bar(symbol, bar)

    @override
    def teardown(self):
        super().teardown()

    @override
    def generateInsights(self, symbol: str):
        # Execute Orders If there should be any
        super().generateInsights(symbol)

    @override
    def executeInsight(self, insight: Insight):
        super().executeInsight(insight)




import abc
import datetime
from typing import Any, Awaitable, Callable, List, Literal, Optional, Union

from typing_extensions import override
import pandas as pd
from dotenv import load_dotenv
from datetime import datetime, timedelta

from .interfaces import IQuote, ISupportedBrokerFeatures, ISupportedBrokers, ITradeUpdate

from .interfaces import IAsset, IAccount, IOrder, IPosition, ITradeUpdateEvent
from ..insight.insight import Insight
from ..utils.timeframe import ITimeFrame, ITimeFrameUnit
from ..strategy.interfaces import IMarketDataStream


class BaseBroker(abc.ABC):
    NAME: ISupportedBrokers = ISupportedBrokers.BASE
    DataFeed: Optional[str]
    PAPER: bool

    supportedFeatures: ISupportedBrokerFeatures

    TICKER_INFO: dict[str, IAsset] = {}

    @abc.abstractmethod
    def __init__(self, name: ISupportedBrokers = ISupportedBrokers.BASE, paper: bool = True, feed: Optional[str] = None) -> None:
        """Abstract class for broker implementations."""
        load_dotenv()
        self.NAME = name
        self.PAPER = paper
        self.DataFeed = feed


    @abc.abstractmethod
    def get_ticker_info(self, symbol: str) -> Union[IAsset, None]:
        pass


    @abc.abstractmethod
    def get_account(self) -> IAccount:
        pass


    @abc.abstractmethod
    def get_position(self, symbol) -> IPosition:
        pass


    @abc.abstractmethod
    def get_positions(self) -> dict[str, IPosition]:
        pass


    @abc.abstractmethod
    def close_position(self, symbol: str, qty: Optional[float] = None, percent: Optional[float] = None) -> Optional[IOrder] :
        """Close a position by symbol"""
        assert qty or percent, "qty or percent must be provided"
        if percent:
            assert (percent > 0) and (percent <= 1), "percent must be with in the range of 0-1"
        pass

    @abc.abstractmethod
    def close_all_positions(self):
        """Close all open positions and cancel all open orders"""
        pass


    @abc.abstractmethod
    def get_orders(self) -> Optional[dict[str, IOrder]]:
        """Get all orders"""
        pass


    @abc.abstractmethod
    def get_order(self, order_id) -> Optional[IOrder]:
        pass


    @abc.abstractmethod
    def get_latest_quote(self, asset: IAsset) -> IQuote:
        pass

    @abc.abstractmethod
    def cancel_order(self, order_id: str) -> Optional[str]:
        pass


    @abc.abstractmethod
    def get_history(self, asset: IAsset, start=(datetime.now() - timedelta(days=7)), end=datetime.now(), resolution=ITimeFrame(5, ITimeFrameUnit.Minute)) -> pd.DataFrame:
        """Get historical data for a given asset open, high, low, close, volume"""

        assert isinstance(
            resolution, ITimeFrame), 'resolution must be of type TimeFrame object'


    @abc.abstractmethod
    def execute_insight_order(self, insight: Insight, asset: IAsset) -> IOrder | None:
        """Manage insight order by planing entry and exit orders for a given insight"""
        assert isinstance(
            insight, Insight), 'insight must be of type Insight object'
        valid, message = insight.validate()
        if not valid:
            raise ValueError("Invalid Entry Insight:", message)
        # assert isinstance(asset, Asset), 'asset must be of type Asset object'


    @abc.abstractmethod
    def startTradeStream(self, callback: Awaitable):
        """Listen to trades and order updates and call the callback function with the data"""
        print("Start Trade Stream -", self.NAME)
        pass


    @abc.abstractmethod
    async def closeTradeStream(self):
        pass


    @abc.abstractmethod
    def streamMarketData(self, callback: Awaitable, assetStreams: List[IMarketDataStream]):
        """Listen to market data and call the callback function with the data"""
        for assetStream in assetStreams:
            assert assetStream['symbol'], 'assetStream must have a symbol'
            assert assetStream['time_frame'], 'assetStream must have a time_frame'
            assert assetStream['type'], 'assetStream must have a type'
        print("Stream Market Data -", self.NAME)
        pass


    @abc.abstractmethod
    async def closeStream(self,  assetStreams: List[IMarketDataStream]):
        pass


    @abc.abstractmethod
    def format_on_bar(self, bar: Any, symbol: Optional[str] = None) -> Optional[pd.DataFrame]:
        """
        Format stream bar data to { symbol: str, bar: -> open, high, low, close, volume}
        -  (data={}, index=MultiIndex[(str, pd.Timestamp)], columns=['open', 'high', 'low', 'close', 'volume']):
        """
        pass

    @abc.abstractmethod
    def format_on_quote(self, quote: Any) -> IQuote:
        """Format stream quote data"""
        pass

    @abc.abstractmethod
    def format_order(self, order: Any) -> IOrder:
        """Format stream Order data"""
        pass
    
    @abc.abstractmethod
    def format_on_trade_update(self, trade: Any) -> tuple[IOrder, ITradeUpdateEvent]:
        """Format stream Trade Order data and event"""
        if isinstance(trade, ITradeUpdate):
            return trade.order, trade.event
        
    @property
    def get_current_time(self) -> datetime:
        """Get the current broker time"""
        return datetime.now()



from ..base_executor import BaseExecutor
from ...insight import InsightState


class DefaultOnClosedExecutor(BaseExecutor):
    """
    ### Executor for Default On Closed
    This executor is used to delete the insight from the strategy when it is closed.

    :param strategy (BaseStrategy): The strategy instance
    
    Author:
        @isaac-diaby
    """

    def __init__(self, strategy):
        super().__init__(strategy, InsightState.CLOSED, "1.0")

    def run(self, insight):
        # Set the default state of the insight
        try:

            if insight.close_order_id is not None:
                del self.STRATEGY.insights[insight.INSIGHT_ID]
                return self.returnResults(True, True, "Insight Closed Successfully.")
            else:
                # set the insight back to filled as it has no close order id
                self.STRATEGY.insights[insight.INSIGHT_ID].updateState(
                    InsightState.FILLED, "Insight does not have a close order id but was set to Closed.")
                return self.returnResults(False, True, "Insight does not have a close order id.")

        except Exception as e:
            return self.returnResults(False, False, f"Error deleting insight: {e}")
